package RPM::Source::SpecSection;

use strict;
use warnings;
use Carp;
use RPM::Source::Macros;
use Clone;

our @ISA=qw/Clone/;

our %SECTION=map {$_=>1} qw!
package
description
prep
build
install
check
clean
files
changelog
pre
post
preun
postun
pretrans
posttrans
trigger
triggerprein
triggerin
triggerun
triggerpostun
!;

# used in R::S::C::Generic2ALT; R::S::Editor
our %SCRIPT_SECTION=map {$_=>1} qw!
pre
post
preun
postun
pretrans
posttrans
trigger
triggerprein
triggerin
triggerun
triggerpostun
!;

our %SHARED_SECTION=map {$_=>1} qw!
prep
build
install
check
clean
changelog
!;

# macros that expand to a new section
our %SECTION_MACROS=map {$_=>1} qw!
!;


########################################
# TODO: support for quoting - something like 
# %post -p "/sbin/ldconfig -n /lib"
# see http://lists.mandriva.com/cooker/2011-03/msg00582.php
########################################
# use Text::ParseWords;
# @words=shellwords($header_line);
########################################
# TODO: move header parsing out of new()

# TODO: %trigger(prein)

sub new {
    my $class = shift;
    my %flag;
    my %flagspace;
    my $self= {
	TYPE=> 'package',
	PACKAGE=> '',
	TRIGGERCOND => undef,
	BODYREF=>[],
	FLAGS=> \%flag,
	FLAGSPACE=> \%flagspace,
	@_
    };
    die "Oops! wrong BODYREF arg: should be an array ref" if ref $self->{BODYREF} ne 'ARRAY';
    #confess "Oops! empty BODYREF" if @{$self->{BODYREF}}==0;
    my $TYPE='package';
    my $PACKAGE='';
    my $verbose=$self->{SPEC}->{VERBOSE};
    my $header_idx=&_get_header_idx($self);
    my $header=$self->{BODYREF}->[$header_idx];
    $header=~s/\s*$//;
    # clean up comments
    $header=~s/#.*$//;
    if ($header=~s/^\s*\%(\w+)\s*// && $SECTION{lc($1)}) {
	$TYPE=lc($1);
	while ($header=~s/^(\S+)(\s*)//) {
	    my $opt=$1;
	    my $optspace=$2;
	    if (substr ($opt,0,2) eq '--') {
		$self->{TRIGGERCOND}=$header;
		$header='';
	    } elsif (substr ($opt,0,1) eq '-') {
		$flagspace{$opt}=$optspace;
		$header=~s/^(\S+)\s*//;
		my $newvalue=$1;
		print STDERR "flag: $opt=$newvalue at \%$TYPE\n" if $verbose>3;
		my $oldvalue=$flag{$opt};
		if ($oldvalue) {
		    # note, in fedora multiple -f flags are allowed. Not in ALT.
		    # -f is handled in Generic2ALT, but there may be other dups as well.
		    print STDERR "WARNING: duplicate flag $opt: $opt $newvalue\n" if $opt ne '-f';
		    $flag{$opt}=[$oldvalue] unless ref $oldvalue;
		    push @{$flag{$opt}}, $newvalue;
		} else {
		    $flag{$opt}=$newvalue;
		}
	    # hack for fedora compatibility
	    } elsif ($opt=~/^\%{?(?:common_description|description_common)/) {
		$self->{BODYREF}->[$header_idx]=~s/\Q$opt\E//;
		push @{$self->{BODYREF}},$opt."\n";
	    } else {
		die "Oops! strange section part: $opt $header (package=$PACKAGE)" if $PACKAGE;
		#print STDERR "package: [$opt]\n";
		$PACKAGE=$opt;
	    }
	}
    }
    die "Oops! strange section: both -n $flag{'-n'} and package: $PACKAGE" if $PACKAGE and $flag{'-n'};
    $PACKAGE='-n'.$flagspace{'-n'}.$flag{'-n'} if $flag{'-n'};
    $self->{TYPE}=$TYPE;
    $self->{PACKAGE}=$PACKAGE;
    bless $self, $class;
    print STDERR "SpecSection: created new section \%$self->{TYPE} $self->{PACKAGE}\n" if $verbose>2;
    print STDERR $self->describe() if $verbose>6;
    return $self;
}

sub describe {
    my ($self)=@_;
    &__check_self($self);
    return "SpecSection: type=$self->{TYPE}; package=$self->{PACKAGE}; body: ==>\n".
	join('',@{$self->{BODYREF}})."============== END BODY ====================\n";
}

# support multiple -f flags!
sub get_flag {
    my ($self,$flag)=@_;
    &__check_self($self);
    my $flagdata=$self->{FLAGS}->{$flag};
    return $flagdata unless ref $flagdata;
    return wantarray ? @$flagdata : $flagdata->[0];
}

sub get_type {
    my $self=shift;
    &__check_self($self);
    return $self->{TYPE};
}

sub set_type {
    my ($self,$newtype)=@_;
    &__check_self($self);
    Carp::croak "main section can not change type" if $self->{PACKAGE} eq '' and $self->{TYPE} eq 'package';
    $self->get_bodyref->[$self->_get_header_idx] =~ s/^\s*\%\Q$self->{TYPE}\E/\%$newtype/ or die "Oops! internal: set_type failed";
    $self->{TYPE}=$newtype;
}

sub get_trigger_condition {
    my $self=shift;
    &__check_self($self);
    return $self->{TRIGGERCOND};
}

sub get_extra_qualifier {
    my $self=shift;
    &__check_self($self);
    my $type=$self->{TYPE};
    my $flag=$self->{FLAGS};
    return '-l'.$self->{FLAGSPACE}->{'-l'}.$flag->{'-l'} if $type eq 'description' and $flag->{'-l'};
    return $self->{TRIGGERCOND};
}

sub spec {
    my $self=shift;
    &__check_self($self);
    return $self->{SPEC};
}

sub set_raw_package {
    my ($self,$newpkg)=@_;
    &__check_self($self);
    Carp::croak "main section can not change packages" if $self->{PACKAGE} eq '' and $self->{TYPE} eq 'package';
    $self->get_bodyref->[$self->_get_header_idx]=~s/\Q$self->{PACKAGE}\E/$newpkg/ or die "Oops! internal: set_package failed";
    $self->{PACKAGE}=$newpkg;
}

sub get_raw_package {
    my $self=shift;
    &__check_self($self);
    return $self->{PACKAGE};
}

sub get_canonical_package {
    my $self=shift;
    &__check_self($self);
    return &RPM::Source::Editor::_canonical_package($self->spec, $self->get_raw_package());
}

sub get_package_name {
    my $self=shift;
    &__check_self($self);
    return &RPM::Source::Editor::_package_name($self->spec, $self->get_raw_package());
}

sub prev {
    my $self=shift;
    &__check_self($self);
    return $self->{'-prev'};
}

sub next {
    my $self=shift;
    &__check_self($self);
    return $self->{'-next'};
}

# TODO: move at Editor.pm level
# prototype
sub _new_section {
    my $self=shift;
    &__check_self($self);
    my @body = split (/(?<=\n)/,@_);
    return RPM::Source::SpecSection->new(
	SPEC=>$self->spec(),
	BODYREF=>\@body);
}
# prototype
sub _append_section {
    my ($self,$new_section)=@_;
    &__check_self($self);
    $new_section->__remove_section_from_chain();
    $new_section->{'-prev'}=$self;
    my $next=$self->{'-next'};
    $self->{'-next'}=$new_section;
    $next->{'-prev'}=$new_section if $next;
    return $new_section;
}
# prototype
sub _prepend_section {
    my ($self,$new_section)=@_;
    &__check_self($self);
    $new_section->__remove_section_from_chain();
    $new_section->{'-next'}=$self;
    my $prev=$self->{'-prev'};
    $self->{'-prev'}=$new_section;
    $prev->{'-next'}=$new_section if $prev;
    return $new_section;
}
# prototype
sub _append_as_section {
    my $self=shift;
    &__check_self($self);
    my $new_section=$self->_new_section(@_);
    $self->_append_section($new_section);
    return $new_section;
}
# prototype
sub __prepend_as_section {
    my $self=shift;
    &__check_self($self);
    my $new_section=$self->_new_section(@_);
    $self->_prepend_section($new_section);
    return $new_section;
}
sub __remove_section_from_chain {
    my $self=shift;
    &__check_self($self);
    my $prev=$self->{'-prev'};
    my $next=$self->{'-next'};
    #confess 'Oops: main section is to be deleted;' if not $prev and $next;
    $next->{'-prev'}=$prev if $next;
    $prev->{'-next'}=$next if $prev;
}

sub delete {
    my $self=shift;
    &__check_self($self);
    my $prev=$self->{'-prev'};
    my $next=$self->{'-next'};
    confess 'Oops: main section is to be deleted;' if not $prev and $next;
    $next->{'-prev'}=$prev if $next;
    $prev->{'-next'}=$next if $prev;
    push @{$prev->get_bodyref}, &__kill_all_exept_macro_conditionals(@{$self->get_bodyref()});
    $prev->set_body(&__filter_empty_macro_if_endif($prev->get_bodyref()));
    $self->__destroy_self;
}

sub is_section {
    my ($self,$type,$pkg, $extra_qualifier)=@_;
    &__check_self($self);
    Carp::croak "looking for section: type not specified." unless $type;
    return 0 if $self->get_type() ne $type;

    # rpm silently ignore section in something like %changelog -n name
    return 1 if $RPM::Source::SpecSection::SHARED_SECTION{$type};

    $pkg='' if (!defined $pkg);
    return 0 if not $self->get_package_name() eq &RPM::Source::Editor::_package_name($self->spec, $pkg);
    return &__are_qualifiers_equal($self->get_extra_qualifier(),$extra_qualifier);
}

sub __are_qualifiers_equal {
    my ( $cond1, $cond2)=@_;
    return 1 if !defined $cond1 && !defined $cond2;
    return 0 if !defined $cond1 || !defined $cond2;
    $cond1=~s/\s+/ /g;
    $cond1=~s/\s*$//;
    $cond1=~s/^\s*//;
    $cond2=~s/\s+/ /g;
    $cond2=~s/\s*$//;
    $cond2=~s/^\s*//;
    return 1 if $cond1 eq $cond2;
    return 0;
}

sub _get_header_idx {
    return 0;
}

sub get_bodyref {
    my $self=shift;
    &__check_self($self);
    return $self->{BODYREF};
}

sub set_body {
    my $self=shift;
    my $ptr=shift;
    &__check_self($self);
    die "not an array pointer!" if ref $ptr ne 'ARRAY';
    foreach (@$ptr) {
	Carp::carp "set_body: got undefined value!" unless defined $_;
    }
    $self->{BODYREF}=$ptr;
}

sub apply_body {
    my ($self, $filtersub, $message) = @_;
    &__check_self($self);
    unless ($message) {
    	my ($package, $filename, $line) = caller;
    	$message="subroutine in $filename at line $line";
    }
    my $is_applied = 0;
    my @tail=@{$self->get_bodyref};
    my @body=&__shift_header_from_a_body_array($self,\@tail);

    # move pushed item(s) before %if* block of next section
    while (@tail>0) {
	my $oldline=shift @tail;
	my @newline=&$filtersub($oldline);
	push @body, @newline;
	$is_applied=1 if @newline>1 or $newline[0] ne $oldline;
    }
    $self->set_body(\@body);
    if ($is_applied) {
	$self->spec->_log_applied("apply_body: [$message]");
    } else {
	$self->spec->_warn_notapplied("apply_body: not applied [$message]");
    }
    return $is_applied;
}

sub visit_body {
    my ($self, $filtersub) = @_;
    &__check_self($self);
    my @tail=@{$self->get_bodyref};
    &__shift_header_from_a_body_array($self,\@tail);
    local $_;
    while (@tail>0) {
	# use global $_ to communicate
	$_=shift @tail;
	&$filtersub($_);
    }
    return;
}

sub map_body {
    my ($self, $filtersub, $message) = @_;
    &__check_self($self);
    unless ($message) {
    	my ($package, $filename, $line) = caller;
    	$message="subroutine in $filename at line $line";
    }
    my $is_applied = 0;
    my @tail=@{$self->get_bodyref};
    my @body=&__shift_header_from_a_body_array($self,\@tail);
    local $_;
    while (@tail>0) {
	my $oldline=shift @tail;
	# use global $_ to communicate
	$_=$oldline;
	&$filtersub($oldline);
	push @body, split (/(?<=\n)/,$_) if defined $_;
	$is_applied=1 if not defined($_) or $_ ne $oldline;
    }
    $self->set_body(\@body);
    if ($is_applied) {
	$self->spec->_log_applied("map_body: [$message]");
    } else {
	$self->spec->_warn_notapplied("map_body: not applied [$message]");
    }
    return $is_applied;
}

sub __shift_header {
    my $self=shift;
    return __shift_header_from_a_body_array($self,$self->{BODYREF});
}

sub __shift_header_from_a_body_array {
    my ($self, $bodyref)=@_;
    &__check_self($self);
    return if $self->is_main_section();
    return shift @$bodyref;
}

sub unshift_body {
    my $self=shift;
    &__check_self($self);
    my @header = &__shift_header($self);
    unshift @{$self->{BODYREF}}, @header, map {split (/(?<=\n)/,$_)} @_;
}

sub unshift_body_before_section {
    my $self=shift;
    &__check_self($self);
    unshift @{$self->{BODYREF}}, map {split (/(?<=\n)/,$_)} @_;
}

sub unshift_body_before {
    my ($self,$pattern,@toinsert) = @_;
    &__check_self($self);
    my $is_applied = 0;
    my @tail=@{$self->get_bodyref};
    my @body=&__shift_header_from_a_body_array($self,\@tail);
    # move pushed item(s) before %if* block of next section
    while (@tail>0 and $tail[0]!~/$pattern/ # condition for add
	) {
	push @body, shift @tail;
    }
    if (@tail) {
	push @body, map { split (/(?<=\n)/,$_) } @toinsert, @tail;
	$is_applied = 1;
    }
    $self->set_body(\@body);
    if ($is_applied) {
	$self->spec->_log_applied("inserted [@toinsert]");
    } else {
	$self->spec->_warn_notapplied("not inserted [@toinsert] - pattern [$pattern] not found");
    }
    return $is_applied;
}

sub unshift_body_after {
    my ($self,$pattern,@toinsert) = @_;
    &__check_self($self);
    my $is_applied = 0;
    my @tail=@{$self->get_bodyref};
    my @body=&__shift_header_from_a_body_array($self,\@tail);
    while (@tail>0 and $tail[0]!~/$pattern/ # condition for add
	) {
	push @body, shift @tail;
    }
    if (@tail) {
	push @body, shift @tail, map { split (/(?<=\n)/,$_) } @toinsert, @tail;
	$is_applied = 1;
    }
    $self->set_body(\@body);
    if ($is_applied) {
	$self->spec->_log_applied("inserted [@toinsert]");
    } else {
	$self->spec->_warn_notapplied("not inserted [@toinsert] - pattern [$pattern] not found");
    }
    return $is_applied;
}

sub push_body {
    my $self=shift;
    &__check_self($self);
    my $bodyptr=$self->{BODYREF};
    # move pushed item(s) before %if* block of next section
    my @tail;
    my $i=$#{$bodyptr};
    while ($i>=0 and
	   # empty line or condition for the next section or comment;
	   # note: |\%else|\%endif is not skipped, as the line should not go into conditonal.
	   # The logic should be somehow rewritten using concept of line visibility.
	   $bodyptr->[$i]=~/^\s*(?:$|#|\%if)/ 
	) {
	unshift @tail, pop @{$bodyptr};
	$i--;
    }

    # look like multline define; to be safe
#%define common_description The Automated Testing Framework (ATF) is a collection
# of libraries and \
#utilities designed to ease unattended application testing in the hands of \
#developers and end users of a specific piece of software. \
# if we add at slash, it will be concatenated.
    if ($bodyptr->[$#{$bodyptr}]=~/\\$/) {
	# TODO: test for actual define here
	push @{$bodyptr}, "\n";
    }

    push @{$bodyptr}, map {split (/(?<=\n)/,$_)} @_;
    push @{$bodyptr}, @tail;
}

sub push_body_after {
    my ($self,$pattern,@toinsert) = @_;
    &__check_self($self);
    my $bodyptr=$self->{BODYREF};
    my $is_applied = 0;
    my @tail;
    my $i=$#{$bodyptr};
    while ($i>=0 and $bodyptr->[$i]!~/$pattern/ # condition for add
	) {
	unshift @tail, pop @{$bodyptr};
	$i--;
    }
    if (@$bodyptr) {
	push @{$bodyptr}, map { split (/(?<=\n)/,$_) } @toinsert, @tail;
	$is_applied = 1;
    }
    if ($is_applied) {
	$self->spec->_log_applied("inserted [@toinsert]");
    } else {
	$self->spec->_warn_notapplied("not inserted [@toinsert] - pattern [$pattern] not found");
    }
    return $is_applied;
}

sub subst_after {
    my $self=shift;
    my $afterpattern = shift;
    my $whatpattern=shift;
    my $replace=shift;
    &__check_self($self);
    my $is_applied = 0;
    my $is_found_after=0;
    my @body=();
    my $count = 0;
    my $maxcount=1;
    my @tail=@{$self->{BODYREF}};
    while (@tail>0 and $tail[0]!~/$afterpattern/ # condition for add
	) {
	push @body, shift @tail;
    }
    $is_found_after = 1 if @tail;
    while (@tail>0 && $count <= $maxcount) {
	my $line=shift @tail;
	#print STDERR "try:$line";
	if ($line=~s/$whatpattern/$replace/g) {
	    $is_applied = 1;
	    $count++;
	}
	push @body, $line;
    }
    push @body, @tail;
    $self->set_body(\@body);
    $is_found_after || $self->spec->_warn_notapplied("pattern [$afterpattern] not found");
    $is_applied || $self->spec->_warn_notapplied("not replaced [$whatpattern] after [$afterpattern]");
    $is_applied && $self->spec->_log_applied("replaced [$whatpattern] after [$afterpattern]");
    return $is_applied;
}

sub empty_body {
    my $self=shift;
    &__check_self($self);
    my @newbody;
    push @newbody, &__shift_header($self);
    push @newbody, __kill_all_exept_macro_conditionals(@{$self->{BODYREF}});
    $self->{BODYREF}=\@newbody;
}

#sub behead {
#    my $self=shift;
#    &__check_self($self);
#    my @newbody=@{$self->{BODYREF}}[1..$#{$self->{BODYREF}}];
#    $self->{BODYREF}=\@newbody;
#}

sub is_empty {
    my $self=shift;
    &__check_self($self);
    return 1 if $#{$self->{BODYREF}}<1;
    return 0 if $self->{BODYREF}->[$self->_get_header_idx]=~/^\%(post|pre|trigger).*\s+-p\s+\S/;
    foreach (@{$self->{BODYREF}}[1..$#{$self->{BODYREF}}]){
	#print "is_empty: [$_]" if ! /^\s*(#.*)?$/;
	return 0 if ! /^\s*(?:(?:#|\%if|\%else|\%endif).*)?$/;
    }
    return 1;
}

sub is_main_section {
    my $self=shift;
    &__check_self($self);
    return 1 if $self->{TYPE} eq 'package' && $self->{PACKAGE} eq '';
    return 0;
}

sub exclude_body {
    my $self=shift;
    my $pattern=shift;
    &__check_self($self);
    Carp::croak "exclude: got undefined argument" unless defined $pattern;
    my @out = &__shift_header($self);
    my $is_applied = 0;
    foreach (@{$self->get_bodyref()}) {
	unless (/$pattern/) {
	    push @out, $_;
	} else {
	    $is_applied++;
	}
    }
    $self->set_body(\@out);
    $self->spec->_report_applied($is_applied,"exclude [$pattern]");
    return $is_applied;
}

sub multi_exclude_body {
    my ($self,@pattern)=@_;
    &__check_self($self);
    Carp::croak "exclude_multi: not enough arguments" unless scalar @pattern;
    my @out = &__shift_header($self);
    my $is_applied = 0;
    my $i = 0;
    my @buffer;
    foreach my $line (@{$self->get_bodyref()}) {
	if ($line=~/$pattern[$i]/) {
	    push @buffer, $line;
	    if (++$i==scalar @pattern) {
		$#buffer=-1;
		$i=0;
		$is_applied++;
	    }
	} else {
	    push @out, @buffer, $line;
	    $#buffer=-1;
	    $i=0;
	}
    }
    $self->set_body(\@out);
    $self->spec->_report_applied($is_applied,"exclude_multi [@pattern]");
    return $is_applied;
}

sub subst_body {
    my $self=shift;
    my $pattern=shift;
    my $replace=shift;
    &__check_self($self);
    Carp::croak "subst: got undefined argument" unless defined $pattern and defined $replace;
    my $is_applied = 0;
    my @out = &__shift_header($self);
    foreach (@{$self->get_bodyref()}) {
	$is_applied++ if s/$pattern/$replace/g;
	push @out, $_;
    }
    $self->set_body(\@out);
    $self->spec->_report_applied($is_applied,"subst [$pattern]");
    return $is_applied;
}

sub subst_body_if {
    my $self=shift;
    my $pattern=shift;
    my $replace=shift;
    my $ifpattern=shift;
    &__check_self($self);
    Carp::croak 'if_pattern is required' unless $ifpattern;
    my @out = &__shift_header($self);
    my $is_applied = 0;
    foreach (@{$self->get_bodyref()}) {
	if (/$ifpattern/) {
	    $is_applied++ if s/$pattern/$replace/g;
	}
	push @out, $_;
    }
    if ($is_applied) {
	$self->spec->_log_applied("subst [$pattern] if [$ifpattern]");
    } else {
	$self->spec->_warn_notapplied("not subst [$pattern] if [$ifpattern]");
    }
    $self->set_body(\@out);
    return $is_applied;
}

sub match_body {
    my $self=shift;
    my $pattern=shift;
    &__check_self($self);
    foreach (@{$self->get_bodyref()}) {
	return 1 if /$pattern/;
    }
    return 0;
}

# compatibility match alias
sub match {
    Carp::cluck "match() method is deprecated. Use match_body()";
    return &match_body;
}

# TODO: drop in next version
sub exclude_multi {
    Carp::cluck "exclude_multi() method is deprecated. Use multi_exclude_body()";
    return &multi_exclude_body;
}

# TODO: drop in next version
sub get_package {
    Carp::cluck "get_package() method is deprecated. Use get_raw_package()";
    return &get_raw_package;
}

{ no warnings;
  *subst=*subst_body;
  *subst_if=*subst_body_if;
  *exclude=*exclude_body;
  *multi_exclude=*multi_exclude_body;
}

sub get_tag {
    my $self=shift;
    my $tag=shift;
    &__check_self($self);
    my %options = (
	RAW=>0,
	@_);
    my ($tag_re,$lctag)=&__get_tag_re_lc($tag);

    my $false_val;
    # TODO: calculate and store context of each section
    # tmp hack - assume section is always true
    my @macro_stack;
    my $macros=$self->spec->macros();
    foreach my $line (@{$self->get_bodyref()}) {
	# %if 0%{?svn_checkout}
	if ($line=~/^\s*\%if\s+(\S+)/) {
	    my $expr=$1;
	    &__push_ifmacro_context(\@macro_stack,$macros->macro_subst($expr));
	} elsif ($line=~/^\s*\%else/) {
	    &__invert_ifmacro_context(\@macro_stack);
	} elsif ($line=~/^\s*\%endif/) {
	    &__pop_ifmacro_context(\@macro_stack);
	} elsif ($line=~/^$tag_re\s*:\s*(.+)$/i) {
	    my $val=$1;
	    $val=~s/\s*$//;
	    #warn "tag=$tag raw =$val";
	    if (!$options{RAW}) {
		if ($val=~/\%\{?$lctag\}?/) {
		    # something like Release: alt1_%{release}1.7
		    # hack; assume %release is not overriden in subpackages
		    my $newval=$macros->{$lctag};
		    $val=$newval if (defined $newval);
		} else {
		    $val=$macros->macro_subst($val);
		}
	    }
	    if (&__get_ifmacro_context(\@macro_stack)) {
		return $val;
	    } else {
		$false_val=$val;
	    }
	}
    }
    return $false_val if defined $false_val;
    return;
}

sub __get_ifmacro_context {
    my ($stack)=@_;
    return $stack->[$#{$stack}] if @$stack>1;
    return 1;
}

sub __push_ifmacro_context {
    my ($stack,$expr)=@_;
    push @$stack,$expr;
}

sub __pop_ifmacro_context {
    my ($stack)=@_;
    pop @$stack;
}

sub __invert_ifmacro_context {
    my ($stack)=@_;
    my $idx=$#{$stack};
    $stack->[$idx]=not $stack->[$idx]  if $idx>-1;
}

sub set_tag {
    my ($self, $tag, $val)=@_;
    &clear_tag($self, $tag) unless defined $val;
    &__check_self($self);
    my $specptr=$self->get_bodyref();
    my $found=0;
    my ($tag_re,$lctag)=&__get_tag_re_lc($tag);
    for (my $i=0; $i<=$#{$specptr}; $i++) {
	if ($specptr->[$i]=~/^$tag_re:(\s*)/i) {
	    $specptr->[$i]="$tag:$1$val\n";
	    $found=1;
	    last;
	}
    }
    $self->unshift_body($tag.': '.$val."\n") unless $found;
    # update macros on set_tag Name,Version,Release
    # TODO: update macros should check tag on set_tag Name,Version,Release,Epoch
    # TODO: do we need macro_subst here ?
    #my $subst_val=$self->spec->macros()->macro_subst($val);
    $self->spec->macros()->{$lctag}=$val;
    my ($urlref, $sourceref, $sourcenum)=&__get_sourcetag_list_num($self,$tag);
    if ($sourceref) {
	my ($prefix,$file)=&__split_url($val);
	$sourceref->[$sourcenum]=$file;
	$urlref->[$sourcenum]=$prefix;
    }
    return $self;
}

sub __split_url {
    my ($url)=@_;
    $url=~s!([^/]+)$!!;
    my $file=$1;
    $file=~s/\s*$//;
    return ($url, $file);
}

sub clear_tag {
    my ($self, $tag)=@_;
    &__check_self($self);
    my ($tag_re,$lctag)=&__get_tag_re_lc($tag);
    my $specptr=$self->get_bodyref();
    for (my $i=0; $i<=$#{$specptr}; $i++) {
	if ($specptr->[$i]=~/^$tag_re:(\s*)/i) {
	    $specptr->[$i]='#'.$specptr->[$i];
	    #last;
	}
    }
    undef $self->spec()->macros()->{$lctag};
    my ($urlref, $sourceref, $sourcenum)=&__get_sourcetag_list_num($self,$tag);
    if ($sourceref) {
	undef $sourceref->[$sourcenum];
	undef $urlref->[$sourcenum];
    }
    return $self;
}

#### protected utilities ####

# TODO: move in some kind of std tools library
### used in JPackage2ALT too :(
sub __filter_empty_macro_if_endif {
    my ($inarrayptr)=@_;
    my @out;
    foreach my $line (@$inarrayptr) {
	push @out, $line;
	if ($line=~/^\s*\%endif/) {
	    my $position_of_if=&__detect_empty_macro_if_endif(\@out);
	    $#out=$position_of_if-1 if $position_of_if >-1;
	}
    }
    return \@out;
}

# returns -1 on failure
# position of %if if found
sub __detect_empty_macro_if_endif {
    my $arrayptr=shift;
    my $is_empty=1;
    my $found_endif=0;
    my $found_else=0;
    my $pos=$#{$arrayptr};
    while ($pos>=0) {
	my $line=$arrayptr->[$pos];
	if ($line=~/^\s*\%endif/) {
	    return -1 if $found_endif;
	    return -1 if $found_else;
	    #print STDERR "found endif $line";
	    $found_endif=1;
	} elsif ($line=~/^\s*\%else/) {
	    return -1 if $found_else;
	    return -1 if ! $found_endif;
	    #print STDERR "found else $line";
	    $found_else=1;
	} elsif ($line=~/^\s*\%if/) {
	    #print STDERR "found pos=$pos\n";
	    return $pos;
	# do not kill comments
	#} elsif ($line!~/^\s*#/) {
	} elsif ($line!~/^\s*$/) {
	    return -1;
	}
	$pos--;
    }

}

#### private utilities ####

sub __get_tag_re_lc {
    my ($tag)=@_;
    my $lctag=lc($tag);
    my $tag_re=$tag;
    $tag_re='source0?' if $lctag eq 'source' or $lctag eq 'source0';
    $tag_re='patch0?' if $lctag eq 'patch' or $lctag eq 'patch0';
    # to support tags like Summary(ru)
    $tag_re=~s,([()]),\\$1,g;
    return $tag_re, $lctag;
}

sub __get_sourcetag_list_num {
    my ($self,$tag)=@_;
    my $sourcenum=0;
    my ($sourceref,$urlref);
    # reread sources on Source\d*, Patch\d*
    if ($tag=~/^Source(\d*)$/i) {
	$sourcenum=$1 if $1;
	$sourceref=$self->spec->raw_sourcelist_ref;
	$urlref=$self->spec->raw_sourceurl_ref;
    } elsif ($tag=~/^Patch(\d*)$/i) {
	my $sourcenum=0;
	$sourcenum=$1 if $1;
	$sourceref=$self->spec->raw_patchlist_ref;
	$urlref=$self->spec->raw_patchurl_ref;
    }
    return $urlref, $sourceref, $sourcenum;
}


## kill all except %if/%else/%endif
sub __kill_all_exept_macro_conditionals {
    my @newbody;
    my @body=map {split /(?<=\n)/} @_;
    while (@body) {
	local $_=shift @body;
	if (/^\s*\%(?:if|else|endif)/) {
	    push @newbody, $_;
	} elsif (/^\s*\%(\S+)/ and $SECTION_MACROS{$1}) {
	    push @newbody, $_;
	    last;
	}
    }
    push @newbody, @body;
    return @newbody;
}

#### private internals ####
sub __check_self {
    my $self=shift;
    Carp::confess ("attempt to manipulate a dead section!") if $self->{DESTROYED};
}

sub __destroy_self {
    my $self=shift;
    %{$self}=(DESTROYED=>1);
}

1;
