package RPM::Source::Tools::SourceAnalyzer;

use Cwd;
use Carp;
use strict;
use warnings;

use File::Path qw(make_path remove_tree);
use File::Temp qw/tempfile tempdir/;
use File::Basename;
use Text::ParseWords;
use DistroMap 0.16;
$DistroMap::Repository::opt_distrodb_version=5;
use Pod::Strip;

our $VERSION=0.005;

# TODO: proper read of m4 args in [...]

our $opt_hint='';
our ($opt_analyze_subdir_autoconf,$opt_lookup_rebased_headers);
our ($opt_resolve_pkgconfig,$opt_allow_unknown_pkgconfig);
our $opt_skip_implicit_pkgconfig=1;
our ($opt_resolve_perl,$opt_allow_unknown_perl);
our $opt_skip_implicit_perl=1;
our $opt_scan_configure_for_broken_cpp_check=0;
our %opt_select_multiple_dependency;

our @LONGOPT=(
    'sourcedep-hint=s'   => \$opt_hint,
    'sourcedep-lookup-rebased-headers'   => \$opt_lookup_rebased_headers,
    'sourcedep-perl-resolve!' => \$opt_resolve_perl,
    'sourcedep-perl-allow-unknown!' => \$opt_allow_unknown_perl,
    'sourcedep-perl-skip-implicit!' => \$opt_skip_implicit_perl,
    'sourcedep-pkgconfig-resolve!' => \$opt_resolve_pkgconfig,
    'sourcedep-pkgconfig-allow-unknown!' => \$opt_allow_unknown_pkgconfig,
    'sourcedep-pkgconfig-skip-implicit!' => \$opt_skip_implicit_pkgconfig,
    'sourcedep-autoconf-search-subdirs!' => \$opt_analyze_subdir_autoconf,
    'sourcedep-select-dependency=s%' => \%opt_select_multiple_dependency,
    );

sub longopt_usage {
    print "SourceAnalyzer Options:
    General Options.
	--sourcedep-hint <hint> - pass a hint to SourceAnalyzer core.
	  current hints are: octave convert
	--sourcedep-autoconf-search-subdirs	search subdirs for extra configure.ac/in files
	--no-sourcedep-autoconf-search-subdirs  negate the option (default)
    Options for special dependencies management.
	Special dependencies are dependencies like type(name) and are often stricter than dependencies on package names. Generation of such dependencies is ruled by options --sourcedep-<type>-*.
	--sourcedep-<type>-resolve		replace all special dependencies with package names.
	--sourcedep-<type>-skip-implicit	drop special dependencies whose packages are mentioned in spec.
	--sourcedep-<type>-allow-unknown	keep all special dependencies, even not found in the repository.

	--sourcedep-pkgconfig-resolve		use package names instead of pkgconfig(name)
	--sourcedep-pkgconfig-skip-implicit	skip pkgconfig(name) dependency if its package is present (default).
	--sourcedep-pkgconfig-allow-unknown	add pkgconfig(name) dependency even if its package is not found.
	--no-sourcedep-pkgconfig-allow-unknown	negate the option (default)
	--no-sourcedep-pkgconfig-resolve	negate the option (default)
	--no-sourcedep-pkgconfig-skip-implicit	negate the option

	--sourcedep-perl-resolve		use package names instead of perl(name)
	--sourcedep-perl-skip-implicit	skip perl(name) dependency if its package is present (default).
	--sourcedep-perl-allow-unknown	add perl(name) dependency even if its package is not found.
	--no-sourcedep-perl-allow-unknown	negate the option (default)
	--no-sourcedep-perl-resolve	negate the option (default)
	--no-sourcedep-perl-skip-implicit	negate the option

	--sourcedep-select-dependency name=package	Select explicit package if the dependency <name> is provided by several packages.

";
}

my %BIN_IGNORE=map {$_=>1} qw!!;
my %BIN_DROP=map {$_=>1} qw!
ar
cc
cl
gcc
gtar
ld
objcopy
ranlib
strip
lib.exe
csc.exe
!;
my %LIB_IGNORE=map {$_=>1} 
# deprecated
qw!
sasl
!, # in std lang environment
qw!
gcc
stdc++
!, #other_unix
'gen', # libgen -_string_pattern-matching_library_Solaris
qw!
getopt
getopt_long
gnugetopt
socket
sunmath
mlib
!, # Must die
qw!
Advapi32
advapi32
bps
clipboard
clntsh
comdlg32
coredll
crypt32
d3d9
d3dcompiler
db2
db2cli
dnsapi
dxguid
gdi32
gds
gds32_ms
iconv
icuin
inet
intl
iphlpapi
kernel32
msimg32
netapi32
odbc32
ole32
oleacc
oleaut32
OpenAL32
openal32
opengl32
posix4
pps
qui
qtest
resmgr
rpcrt4
screen
sendfile
setupapi
slog2
sqldb
user32
WFD
winmm
winspool
wldap32
Ws2_32
ws2_32
ws2
ws32
wsock32
!;
my %HEADER_IGNORE=map {$_=>1} (
'Carbon/Carbon.h',	# MAC
'CoreServices/CoreServices.h', # MAC
'QuickTime/QuickTime.h', # MAC
'fp_class.h',		# *
'ieeefp.h',		# *
'machine/endian.h',	# BSD
'machine/types.h',	# BSD
'nan.h',		# BSD
'sndio.h',		# OpenBSD,...
'sunmath.h',		# Solaris
'sys/byteswap.h',	# SPARC?
'sys/dg_sys_info.h',	# AIX, HP-UX
'sys/audioio.h',	# BSD
'sys/cdio.h',		# BSD
'sys/fbio.h',		# BSD
'sys/disk.h',		# BSD
'sys/endian.h',		# BSD
'sys/extaddr.h',	# BSD
'sys/if.h',		# BSD
'sys/int_types.h',	# BSD
'sys/mode.h',		# AIX
'sys/sockio.h',		# BSD
'sys/scsiio.h',		# BSD
'sys/videoio.h',	# BSD
'sys/systeminfo.h',	# Solaris
'sys/processor.h',	# Solaris
'sys/loadavg.h',	# Solaris
'sys/fpu.h',		# *
'windows.h',	# Must die
'winsock.h',	# Must die
'winsock2.h',	# Must die
'ws2tcpip.h',	# Must die
'basetsd.h',	# Must die
'hash_map',	# Must die
'hash_set',	# Must die
'iphlpapi.h',	# Must die
'tchar.h',	# Must die
'mmsystem.h',	# Must die
'DShow.h',	# DirectShow
'strstream',
),
# extra sys headers
qw!
android/native_window.h
cygwin/fs.h
machine/soundcard.h
sys/acpa.h
sys/apparmor.h
sys/audio.h
sys/bufmod.h
sys/devpoll.h
sys/disklabel.h
sys/dl.h
sys/dlpi_ext.h
sys/dlpi.h
sys/dlpihdr.h
sys/dvdio.h
sys/event.h
sys/extattr.h
sys/fpu.h
sys/ioccom.h
sys/mib.h
sys/mixer.h
sys/mkdev.h
sys/mntent.h
sys/ndd_var.h
sys/net/nit.h
sys/proc.h
sys/sam9407.h
sys/sched.h
sys/scsi.h
sys/scsi/scsi_types.h
sys/sdt.h
sys/stdtypes.h
sys/stropts.h
sys/telephony.h
sys/uc_access.h
sys/ucred.h
sys/utime.h
sys/videodev2.h
!;
my %HEADER_DROP=map {$_=>1} qw!
config.h
!;

my %GIR_IGNORE=map {$_=>1} qw//;
my %PERL_IGNORE=map {$_=>1} qw!
______misc_____________________
perl(Win32/API.pm)
perl(v018.nofile2.pl)
!;
my %PERL_DROP=map {$_=>1} qw!
perl(ExtUtils/MM_AIX.pm)
perl(ExtUtils/MM_BeOS.pm)
perl(ExtUtils/MM_Cygwin.pm)
perl(ExtUtils/MM_Darwin.pm)
perl(ExtUtils/MM_DOS.pm)
perl(ExtUtils/MM_MacOS.pm)
perl(ExtUtils/MM_NW5.pm)
perl(ExtUtils/MM_OS2.pm)
perl(ExtUtils/MM_QNX.pm)
perl(ExtUtils/MM_UWIN.pm)
perl(ExtUtils/MM_VMS.pm)
perl(ExtUtils/MM_VOS.pm)
perl(ExtUtils/MM_Win32.pm)
perl(ExtUtils/MM_Win95.pm)
______generated_modules________
perl(Marpa/XS/Version.pm)
perl(XML/SAX/ExpatXS/Preload.pm)
!;

my %raw_perl_ignore=map {$_=>1} qw/
5
5.005
v5
bytes
constant
feature
inc
integer
lib
locale
strict
utf8
vars
version
warnings
Carp
Cwd
Data::Dumper
File::Basename
File::CheckTree
File::Compare
File::Copy
File::Find
File::Glob
File::Path
File::Spec
File::Temp
Getopt::Long
Getopt::Std
POSIX
Tie::Array
Tie::Handle
Tie::Hash
Tie::Memoize
Tie::Scalar
/;

# TODO
#my %IGNORE=(
#'gir' => \%GIR_IGNORE,
#'devel-libs' => \%LIB_IGNORE,
#'perl' => \%PERL_IGNORE,
#'headers' => \%HEADER_IGNORE,
#'pkg-config' => {},
#    );

my %STORAGE_KEY=(
'gir','girs',
'vapi','vapi',
'devel-libs','libs',
'headers','hdrs',
'headers-rebased','hdrs',
#'path','bin',
'perl','perl',
'pkg-config','pkgconf',
);

&__add_file_to_flag('/usr/share/srpmtools/data/stdheaders.txt',\%HEADER_IGNORE);

my %BIN2PKG=qw!
/usr/bin/c++	gcc-c++
/usr/bin/dvips	texlive-base-bin
!;

# TODO:
my %HEADER2PKG=qw/
jni.h	java-devel
/;

my %NOT_SCRIPTS=map {$_=>1} qw/
ABOUT
API
ARTISTIC
AUTHORS
BUGS
CHANGES
Changes
ChangeLog
Changelog
CHANGELOG
configure
Copying
COPYING
copying
COPYRIGHT
copyright
crosscompile
CTAGS
ETAGS
depcomp
Dependencies
Doxyfile
FAQ
GPL
highscores
install-sh
INSTALL
LGPL
LICENSE
libtool
MACHINES
Makefile
makefile
GNUMakefile
MANIFEST
MANUAL
missing
mkinstalldirs
MPI
NEWS
PKG-INTRO
py-compile
README
requirement
THANKS
TODO
TUTORIAL
SConstruct
shtool
VERSION

/;

sub new {
    my $class = shift;
    my $self = {
	__IS_BINARY__ => 0,
	-verbose => 0,
	-current_buildreq => {},
	-hint => $opt_hint,
	-resolve_pkgconfig => $opt_resolve_pkgconfig,
	-allow_unknown_pkgconfig => $opt_allow_unknown_pkgconfig,
	-skip_implicit_pkgconfig => $opt_skip_implicit_pkgconfig,

	-resolve_perl => $opt_resolve_perl,
	-allow_unknown_perl => $opt_allow_unknown_perl,
	-skip_implicit_perl => $opt_skip_implicit_perl,
	# TODO: something like (rebased filelist) '^examples/'
	#-skip_analyze_by_pattern => $opt_skip_analyze_by_pattern,

	-lookup_rebased_headers => $opt_lookup_rebased_headers,

	-select_multiple_dependency => {},
	-analyze_subdir_autoconf => $opt_analyze_subdir_autoconf,
	-scan_configure_for_broken_cpp_check => $opt_scan_configure_for_broken_cpp_check,

	@_
    };
    bless $self, $class;
    
    my $hint=$self->{-hint};
    if ($hint and $hint eq 'octave') {
	$self->{-analyze_subdir_autoconf}=1 if not defined $self->{-analyze_subdir_autoconf};
    }
    return $self;
}

sub is_arch_dependent {
    my ($self,$flag)=@_;
    return $self->{__IS_BINARY__};
}

sub __set_binary {
    my ($self,$flag)=@_;
    $self->{__IS_BINARY__}=1;
}

sub analyze {
    my ($self,$archive_obj)=@_;
    Carp::cluck "analyze: $archive_obj should be an object of SourceArchive" unless ref $archive_obj;
    my $verbose=$self->{-verbose};
    my $buildreq=$self->{-buildreq_storage};
    my $distroinfo=$self->{-distroinfo};
    Carp::croak "-distroinfo is required" unless $self->{-distroinfo};
    Carp::croak "-buildreq_storage is required" unless $self->{-buildreq_storage};

    my ($lang_c,$lang_cxx,$lang_fortran,$lang_python,$lang_java);
    my ($prog_pod2man,$prog_swig);
    my (@configurac,@makefileam,@qmakepro,@guess_perl,@guess_test_perl,@guess_scripts);
    my (@configureoctave,@cmakelists,@perl_buildscripts,@raw_configure,@pomxml);
    my (@c_files,@cpp_files);

    my $hint=$self->{-hint};
    my $force_perl_by_extensions = ($hint eq 'perl');
    # do not enter git dir
    my @filelist=grep {not m!(?:^|/)\.git(?:/|$)!} @{$archive_obj->filelist()};
    my ($perl_inc_Module_Install,$perl_kill_CPANPLUS);

    foreach (@filelist) {
	if (m!(?:^|/)debian/!) {
	    # skip
	} elsif (m!(?:^|/)configure\.(?:ac|in)$!) {
	    push @configurac, $_;
	} elsif (m!(?:^|/)configure$!) {
	    push @raw_configure, $_;
	} elsif (m!(?:^|/)configure\.(?:base|add)$!) {
	    push @configureoctave, $_;
	} elsif (m!(?:^|/)Makefile\.am$!) {
	    push @makefileam, $_;
	} elsif (m!(?:^|/)(?:Makefile|Build)\.PL$!) {
	    push @perl_buildscripts, $_;
	    $force_perl_by_extensions=1;
	} elsif (m!(?:^|/)inc/Module/AutoInstall\.pm$!) {
	    push @guess_perl, $_;
	    $perl_kill_CPANPLUS=1;
	} elsif (m!(?:^|/)inc/Module/Install\.pm$!) {
	    # TODO:
	    #push @guess_perl, $_;
	    #perl-Module-Install-1.06-alt1.noarch provides perl(inc/Module/Install.pm)
	    $perl_inc_Module_Install=1;
	} elsif (m!\.pr[io]$!) {
	    push @qmakepro, $_;
	} elsif (m!(?:^|/)CMakeLists\.txt$!) {
	    push @cmakelists, $_;
	} elsif (m!(?:^|/)man.*\.pod$!) {
	    $prog_pod2man=1;
	} elsif (m!(?:^|/)__init__\.py$!) {
	    $lang_python=1;
	} elsif (m!\.(?:c|h)$!) {
	    push @c_files, $_;
	} elsif (m!\.i$!) {
	    $prog_swig=1;
	} elsif (m!\.(?:C|CPP|CXX|c(?:c|pp|xx))$!) {
	    # cc|cxx|C|cpp
	    push @cpp_files, $_;
	    $lang_cxx=1;
    # TODO: file' magic, skip if unpack
	} elsif (m!\.p[lm]$!) {
	    push @guess_perl, $_;
	} elsif (m!\.p[lm]\.PL$!) {
	    push @guess_perl, $_;
	} elsif (m!(?:^|/)t/.+\.(?:t|pm)$!) {
	    push @guess_test_perl, $_;
	} elsif (m!\.f?cgi$!) {
	    push @guess_scripts, $_;
	# files in root w/o extension
	} elsif (m!(?:^|/)[^./]+$!) {
	    push @guess_scripts, $_ unless $NOT_SCRIPTS{basename($_)};
	} elsif (m!(?:^|/)pom\.xml$!) {
	    push @pomxml, $_;
	}
    }

    print STDERR "guess_scripts=",join(',',@guess_scripts),"\n" if $verbose >2;

    if ($hint and $hint eq 'octave') {
	unshift @configurac, @configureoctave;
    }
    if (not $self->{-analyze_subdir_autoconf}) {
	# temporary?; let us analyze root only
	@configurac=grep {not m!^.*/.*/configure!} @configurac;
    }

    my $unpack_c=0;
    $lang_cxx=1 if @cmakelists;
    $lang_c=1 if @c_files;
    if ($hint and $hint eq 'octave' or $hint eq 'perl' or $hint eq 'nodejs') {
	$unpack_c=1;
    }
    if ($hint and $hint eq 'octave') {
	$lang_fortran=1 if grep {/\.(?:f)$/} @filelist;
    }

    print STDERR "SourceAnalyzer: perl build scripts found:", join(' ',@perl_buildscripts),"\n" if $verbose >1;

    #$buildreq->add_buildreq(&__perl_pkg2dep('Module::Build')) if grep {m!(?:^|/)Build.PL!} @perl_buildscripts;
    $buildreq->add_buildreq('swig') if $prog_swig;
    $buildreq->add_buildreq('perl-devel') if @perl_buildscripts;
    $buildreq->add_buildreq(&__perl_pkg2dep('CPAN')) if $perl_inc_Module_Install;
    $buildreq->add_buildreq('/usr/bin/pod2man') if $prog_pod2man;
    $buildreq->add_buildreq('gcc-c++') if $lang_cxx;
    $buildreq->add_buildreq('gcc-fortran') if $lang_fortran;
    $buildreq->add_buildreq('python-devel') if $lang_python;
    $self->__set_binary(1) if $lang_c or $lang_cxx or $lang_fortran;

    my $curdir = getcwd;

    push @guess_perl, @guess_test_perl,@perl_buildscripts if @perl_buildscripts;
    my @unpack_list=(@configurac,@makefileam,@qmakepro,@guess_perl,@guess_scripts,@pomxml,@cmakelists);
    push @unpack_list, @raw_configure if $self->{-scan_configure_for_broken_cpp_check};
    push @unpack_list, @c_files, @cpp_files if $unpack_c;

    if (@unpack_list) {
	# unpack 
	my $unpackdir = $archive_obj->extract(-list=>\@unpack_list);

	if (grep {not -e "$unpackdir/$_"} @unpack_list) {
	    # some of unpackef files seems to be (broken) symlinks; no choice, unpack all
	    $unpackdir = $archive_obj->extract();
	}

	foreach my $configuracfile (@configurac) {
	    $self->__analyze_configure_ac("$unpackdir/$configuracfile",$configuracfile);
	}
	if ($self->{-scan_configure_for_broken_cpp_check}) {
	    foreach my $raw_configurefile (@raw_configure) {
		$self->__analyze_raw_configure("$unpackdir/$raw_configurefile",$raw_configurefile);
	    }
	}
	foreach my $amfile (@makefileam) {
	    $self->__analyze_makefile_am("$unpackdir/$amfile",$amfile);
	}
	unless (@configurac) {
	    foreach my $qproject (@qmakepro) {
		$self->__analyze_qmakepro("$unpackdir/$qproject",$qproject);
		# TODO: add proper /.../qmake (qt3,qt4)
	    }
	    foreach my $cmakeproject (@cmakelists) {
		$self->__analyze_cmakelists("$unpackdir/$cmakeproject",$cmakeproject);
	    }
	}
	if (@guess_perl or @guess_scripts) {
	    $self->__analyze_scripts_by_magic($unpackdir,[@guess_scripts,@guess_perl], $force_perl_by_extensions);
	}

	if ($unpack_c and (@c_files or @cpp_files)) {
	    foreach my $cfile (@c_files,@cpp_files) {
		$self->__analyze_c_cpp_headers("$unpackdir/$cfile",$cfile);
	    }
	}
    }

    $buildreq->del_buildreq(&__perl_pkg2dep('CPANPLUS::Backend')) if $perl_kill_CPANPLUS;
    # TODO: interdeps table
    $buildreq->add_buildreq(&__perl_pkg2dep('Moose')) if $buildreq->has_buildreq(&__perl_pkg2dep('Any::Moose'));
    $buildreq->add_buildreq(&__perl_pkg2dep('Test::Most')) if $buildreq->has_buildreq(&__perl_pkg2dep('Test::Class::Most'));

    # TODO: add explicit cleanup interface
    #foreach my $dep ($buildreq->get_list_buildreq()) {
    #	#perl-Module-Install-1.06-alt1.noarch provides perl(inc/Module/Install.pm)
    #   $buildreq->del_buildreq($dep) if $dep=~m!^perl(inc/!;
    #}

}

sub __unique_list {
    my @out;
    my %req;
    foreach my $a (@_) {
	if (not defined $a) {
	    Carp::cluck "undef in __unique_list";
	    push @out, $a;
	} elsif (ref($a)) {
	    push @out, $a;
	} else {
	    $req{$a}=1;
	}
    }
    return @out, keys %req;
}

# sub __unique_and_filter_list {
#     my %ignore=map {$_=>1} @{shift()};
#     my @out;
#     my %req;
#     foreach my $a (@_) {
# 	if (ref($a)) {
# 	    die "Oops: internal error: not an ARRAY" if ref ($a) ne 'ARRAY';
# 	    push @out, [ grep {not $ignore{$_}} @$a];
# 	} elsif (not $ignore{$a}) {
# 	    $req{$a}=1;
# 	}
#     }
#     return @out, keys %req;
# }


sub __analyze_configure_ac {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::Autoconf',
	     -macrohook => \&_autoconf_macro_handler,
	);
}

sub __analyze_makefile_am {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::AutoMake',
	     -varhook => \&_automake_var_handler,
	);
}

sub __analyze_qmakepro {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::QMake',
	     -varhook => \&_qmake_var_handler,
	     -postfilehook => \&_qmake_postfile_handler,
	);
}

sub __analyze_cmakelists {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::CMake',
	     -pathhook => \&_cmake_path_handler,
	     -postfilehook => \&_cmake_postfile_handler,
	);
}

sub __analyze_c_cpp_headers {
    my ($self,$filename,$inputname)=@_;
    &__analyze_file_using($self,$filename,$inputname,
	     'SourceAnalyzer::Language::CPP_header',
	     -headerhook => \&_cpp_header_handler,
	);
}


my %KNOWN_DBD_DRIVERS=map {$_=>1} qw/
DBD::Adabas
DBD::ADO
DBD::Amazon
DBD::AnyData
DBD::ASAny
DBD::Cego
DBD::Chart
DBD::CSV
DBD::DashProfiler
DBD::DB2
DBD::DBM
DBD::DBMaker
DBD::drizzle
DBD::DtfSQLmac
DBD::Empress
DBD::EmpressNet
DBD::ExampleP
DBD::Excel
DBD::File
DBD::Firebird
DBD::Fulcrum
DBD::Gofer
DBD::Google
DBD::Illustra
DBD::Informix
DBD::Informix4
DBD::Ingres
DBD::IngresII
DBD::InterBase
DBD::iPod
DBD::JDBC
DBD::LDAP
DBD::Log
DBD::MaxDB
DBD::Mimer
DBD::Mock
DBD::monetdb
DBD::mSQL
DBD::Multi
DBD::Multiplex
DBD::mysql
DBD::mysqlPP
DBD::NET
DBD::NullP
DBD::ODBC
DBD::Oracle
DBD::Ovrimos
DBD::PassThrough
DBD::Pg
DBD::PgLite
DBD::PgPP
DBD::PgSPI
DBD::Plibdata
DBD::pNET
DBD::PO
DBD::PrimeBase
DBD::Proxy
DBD::QBase
DBD::RAM
DBD::RDB
DBD::RDFStore
DBD::Redbase
DBD::Safe
DBD::Salesforce
DBD::SearchServer
DBD::Simulated
DBD::Solid
DBD::Sponge
DBD::Sprite
DBD::SQLAnywhere
DBD::Sqlflex
DBD::SqlflexTest
DBD::SQLite
DBD::SQLite2
DBD::Sybase
DBD::Sys
DBD::Template
DBD::TemplateSS
DBD::Teradata
DBD::TimesTen
DBD::TreeData
DBD::TSM
DBD::Unify
DBD::Wire10
DBD::WMI
DBD::WTSprite
DBD::XBase
DBD::Yaswi
/;

sub __analyze_scripts_by_magic {
    my ($self, $unpackdir,$filelist, $force_perl_by_extensions)=@_;
    my $storage=__::SourceAnalyzer::Collect->new();
    foreach my $file (@$filelist) {
	my $magic = `file -b "$unpackdir/$file"`;
	if ($magic =~ /^perl script text/ || $magic =~ /Perl5 module source/
	    # see examples-magic
	    || ($magic =~ /ASCII text/ and $file =~ /\.(?:pl)$/)
	    # ASCII English text
	    || ($force_perl_by_extensions and ($file =~ /\.(?:p[ml]|PL)$/ or $file =~ m!(?:^|/)t/.+\.t$!))
) {
	    &__skim_perl_file($self,"$unpackdir/$file",$storage);
	} else {
	    print STDERR "skipped $file: magic=$magic\n" if $self->{-verbose} >2;
	}
    }
    my $inputname='script parser';
    $self->__analyze_storage($storage,$inputname);
}

sub __skim_perl_file {
    my ($self,$file,$storage)=@_;
    # &__analyze_file_using($self,$filename,$inputname,
    # 	     'SourceAnalyzer::Language::Perl',
    # 	);
    my $base_multiline;

    # generated files (\.PL) and files that are 'use'd. but don't have its own package
    if ($file=~m!lib/(.*\.pm)(?:\.PL)?$!) {
	push @{$storage->{perl_ignore}}, &__perl_req2dep($1);
    }


    my $pod_stripper=Pod::Strip->new;              # create parser
    my $podless;                        # set output string
    $pod_stripper->output_string(\$podless);       # see Pod::Simple

    open my $fh, "<", $file or Carp::confess($!);
    $pod_stripper->parse_file($fh);
    close $fh;

    my $line_count=0;
    local $_;
    while (length($podless)) {
	if ($podless=~s/^([^\n]*)\n//) {
	    $_=$1;
	} else {
	    $_=$podless;
	    $podless='';
	}
	unless ($line_count) {
	    ##!/usr/bin/perl -MDevel::Cover
	    if (s,^#!/usr/bin/perl\s*,,) {
		my @mopts=grep{/^-M/} split(/\s+/,$_);
		map {s,-M,,} @mopts;
		map {push @{$storage->{perlreq}}, &__perl_pkg2dep($_)} @mopts;
	    }
	}
	$line_count++;
	last if /^\s*__END__\s*;?\s*$/;

	if (defined $base_multiline) {
	    if (/^([^;]*);/) {
		$base_multiline.=' '.$1;
		&__perl_process_use_base($base_multiline);
		$base_multiline=undef;
	    } else {
		$base_multiline.=' '.$_;
	    }
	    next;
	}

	# experimental.
	if (/[\W]dbi:([a-zA-Z]\w*):/ and $KNOWN_DBD_DRIVERS{'DBD::'.$1}) {
	    push @{$storage->{perlreq}}, &__perl_pkg2dep('DBD::'.$1);
	}

	if (/^\s*use\s+([\w\d_:]+)/ and not $raw_perl_ignore{$1}) {
	    my $module=$1;
	    if ($module eq 'base' or $module eq 'parent') {
		push @{$storage->{perlreq}}, &__perl_pkg2dep($module);
		my $base=$_;
		$base=~s/^\s*use\s+(?:base|parent)\s+//;
		if ($base=~s/\s*;\s*$//) {
		    &__perl_process_use_base($base,$storage);
		} else { 
		    $base_multiline=$_;
		    next;
		}
	    } else {
		# something like pod - still slips in :(
		# use a C<tie>-based interface to use the standard Perl I/O operators.
		push @{$storage->{perlreq}}, &__perl_pkg2dep($module) if not /^\s*use\s+[\w\d_:]+\s+\w+\s+\w+\s/;
	    }

	# with:
	# with qw(Data::Stream::Bulk); see Data-Stream-Bulk-AnyEvent-v0.0.2
	} elsif (m/^\s*with\s*qw['\[\]!(){}|\/]([\w\d_:\s]+)['\[\]!(){}|\/]\s*;\s*$/) {
	    push @{$storage->{perlreq}}, map{&__perl_pkg2dep($_)} split (/\s+/,$1);
	} elsif (m/^\s*with\s*(?:\(\s*)?['"]([^\)]*)['"](?:\)\s*)?\s*;\s*$/) {
	    push @{$storage->{perlreq}}, map{&__perl_pkg2dep($_)} split (/['"]\s*,\s*['"]/,$1);

	} elsif (m!^\s*use_ok\s*\(\s*['"]([\w_:]+)['"]\s*\)\s*;!) {
	    my $module=$1;
	    push @{$storage->{perlreq}}, &__perl_pkg2dep($module);
	} elsif (m!(?:^|\s|\{|;)require\s+['"]([\w\d_/\.]+)['"]\s*;!) {
	    # TODO: CHECK If softreq (prev.line is eval {)
	    push @{$storage->{perlreq}}, &__perl_req2dep($1);
	} elsif (m!(?:^|\s|\{|;)require\s+([\w:]+)\s*;!) {
	    # TODO: CHECK If softreq (prev.line is eval {)
	    push @{$storage->{perlreq}}, &__perl_pkg2dep($1);
	} elsif (/^\s*package\s+([\w\d_:]+)/) {
	    my $module=$1;
	    push @{$storage->{perl_ignore}}, &__perl_pkg2dep($module) if not /^\s*package\s+[\w\d_:]+\s+\w+\s+\w+\s/;
	} elsif (/^__(?:END|DATA)__/) {
	    last;
	}
    }
}

sub __perl_process_use_base {
    my ($base,$storage)=@_;
    # clean up comments
    $base=~s/;\s*#.*$//;
    # use base qw/Class::Accessor/;
    if ($base=~m/^qw['\[\]!(){}|\/]([\w\d_:\s]+)['\[\]!(){}|\/]/) {
	push @{$storage->{perlreq}}, map{&__perl_pkg2dep($_)} split (/\s+/,$1);
	# use base 'Class::Accessor';
    } elsif ($base=~m/^(?:\(\s*)?['"]([^\)]*)['"]\s*(?:\)\s*)?$/) {
	push @{$storage->{perlreq}}, map{&__perl_pkg2dep($_)} split (/['"]\s*,\s*['"]/,$1);
	# use base Class::Accessor;
    } elsif ($base=~m/^([\w\d_:]+)/) {
	push @{$storage->{perlreq}}, map{&__perl_pkg2dep($_)} $1;
    } else {
	print STDERR "TODO:SourceAnalyzer:use_base: can't parse [$base]:$_\n";
    }
}

# TODO: the same sub is used in PerlBuilder -- share it
sub __perl_pkg2dep {
    my $dep="perl($_[0].pm)";
    $dep=~s!::!/!g;
    return $dep;
}
sub __perl_req2dep {
    my $dep="perl($_[0])";
    return $dep;
}

sub __analyze_raw_configure {
    my ($self,$filename,$inputname)=@_;
    open my $fh, "<", $filename or die $!;
    local $_;
    while (<$fh>) {
	chomp;
	if (/\$LINENO: error: C\+\+ preprocessor \\"\$CXXCPP\\" fails sanity check/) {
	    my $storage=__::SourceAnalyzer::Collect->new();
	    push @{$storage->{pkgsreq}},'gcc-c++';
	    $inputname||='broken C++ test in configure';
	    $self->__analyze_storage($storage,$inputname);
	    last;
	}
    }
    close $fh;
}


sub __analyze_file_using {
    my ($self,$filename,$inputname,$class,@args)=@_;
    my $storage=__::SourceAnalyzer::Collect->new();
    my $parser=$class->new(
	-storage => $storage,
	-verbose => $self->{-verbose},
	@args,
	);
    $parser->parse_file($filename);
    $inputname||=basename($filename);
    $self->__analyze_storage($storage,$inputname);
}


sub __analyze_storage {
    my ($self,$storage,$inputname)=@_;
   
    my $verbose=$self->{-verbose};
    my $buildreq=$self->{-buildreq_storage};
    my $distroinfo=$self->{-distroinfo};
    my $current_buildreq=$self->{-current_buildreq};

    &__uniquify_storage($storage);
    &__debug_storage($storage,$inputname.' (uniquified)') if $verbose>1;

    my @IGNORE=@{$storage->{pkgs_ignore}};
    my @BR=@{$storage->{pkgsreq}};
    my $flag_not_found;
    &__add_br2($self,\@BR,\@IGNORE,$current_buildreq,$distroinfo,'devel-libs',$storage,\$flag_not_found,\%LIB_IGNORE,{});
    &__add_br2($self,\@BR,\@IGNORE,$current_buildreq,$distroinfo,'headers',$storage,\$flag_not_found,\%HEADER_IGNORE,\%HEADER_DROP);
    &__add_br2($self,\@BR,\@IGNORE,$current_buildreq,$distroinfo,'gir',$storage,\$flag_not_found,\%GIR_IGNORE,{});

    &__add_resolvable_br2($self,\@BR,\@IGNORE,'pkgconfig',sub { return "pkgconfig($_[0])"},$current_buildreq,$distroinfo,'pkg-config',$storage,\$flag_not_found,{},{});
    &__add_resolvable_br2($self,\@BR,\@IGNORE,'perl',sub { return $_[0]},$current_buildreq,$distroinfo,'perl',$storage,\$flag_not_found,\%PERL_IGNORE,\%PERL_DROP);

    if ($self->{-lookup_rebased_headers}) {
	# TODO:
	#&__add_br2($self,\@BR,\@IGNORE,$current_buildreq,$distroinfo,'headers-rebased',$storage,\$flag_not_found,\%HEADER_IGNORE,\%HEADER_DROP);
    }


  PROG:
    foreach my $name (@{$storage->{pathreq}}) {
	my @names=($name);
	@names=(@$name) if ref $name;
	# like "smth --config"
	@names=map {s/^["']//;s/["']$//;s/\s.*$//;$_} @names;
	my @data;
	my $file;
      CANDIDATE:
	foreach my $candidate (@names) {
	    $candidate=~s/\$EXEEXT$//;
	    next PROG if $BIN_DROP{$candidate};
	    foreach my $prefix (qw!/usr/bin/ /usr/sbin/!) {
		my $file0=$prefix.$candidate;
		my @data0=$distroinfo->translate('bin',$file0);
		# for alternatives in /usr/bin
		push @data0, $distroinfo->translate('provides',$file0) unless @data0;
		next PROG if @names>1 and (&__hash_have_any($current_buildreq,\@data0) or $current_buildreq->{$file0});
		if(@data0) {
		    $file=$file0;
		    @data=@data0;
		    last CANDIDATE;
		}
	    }
	    foreach my $prefix (qw!/bin/ /sbin/!) {
		my $file0=$prefix.$candidate;
		my @data0=$distroinfo->translate('bin',$file0);
		next PROG if @data0; # /bin utilities are considered as always present (basesytem)
	    }
	}
	if (@data>1) {
	    next PROG if &__hash_have_any($current_buildreq,\@data) or $current_buildreq->{$file};
	    if ($BIN2PKG{$file}) {
		&__push(\@BR, $BIN2PKG{$file});
	    } elsif ($self->{-select_multiple_dependency}->{$file}) {
		&__push(\@BR, $self->{-select_multiple_dependency}->{$file});
	    } elsif ($opt_select_multiple_dependency{$file}) {
		&__push(\@BR, $opt_select_multiple_dependency{$file});
	    } else {
		warn "INFO: SourceAnalyzer: multiple provides in path for $file: ",join(' ',@data),"\n";
		&__push(\@BR, $file);
	    }
	} elsif (not @data and &__names_are_not_ignored(\%BIN_IGNORE,\%BIN_DROP,@names)) {
	    warn "INFO: SourceAnalyzer: can't find ".join(' ',@names)." in /usr/bin /usr/sbin.\n";
	    $flag_not_found=1;
	} else {
	    if ($BIN2PKG{$file}) {
		&__push(\@BR, $BIN2PKG{$file});
	    } else {
		&__push(\@BR, $file);
	    }
	}
    }
    
    @BR = sort &__unique_list(@BR);
    print STDERR "$inputname: BR(detected):", join(' ',@BR),"\n" if $verbose >1;
    $buildreq->add_buildreq_for("guessed from $inputname: ",@BR);

    @IGNORE = sort &__unique_list(@IGNORE);
    print STDERR "$inputname: Ignore BR:", join(' ',@IGNORE),"\n" if $verbose >1;
    $buildreq->ignore_buildreq(@IGNORE);
    
    my $lang=$storage->{lang};
    $buildreq->add_buildreq('gcc-c++') if $lang->{cxx};
    $buildreq->add_buildreq('gcc-fortran') if $lang->{fortran};
    $buildreq->add_buildreq('python-devel') if $lang->{python};

    # TODO: 'convert' hint?
    # broken impls of AC_PROG_CC might require g++ to be installed :( see libmodelfile

    warn "SourceAnalyzer: some deps weren't found. Is DistroMap database outdated?\n" if $flag_not_found;
}

sub __add_br2 {
    my ($self,$BR,$IGNORE,$current_buildreq,$distroinfo,$db,$storage,$flag_not_found_ref,$ignore,$input_drop)=@_;
    my $key=$STORAGE_KEY{$db};
    die "internal error: no storage key for $db" unless $key;
    &__add_br($self,$IGNORE,$current_buildreq,$distroinfo,$db,$storage->{$key.'_ignore'},undef,{},{});
    &__add_br($self,$BR,$current_buildreq,$distroinfo,$db,$storage->{$key.'req'},$flag_not_found_ref,$ignore,$input_drop);
}

sub __add_br {
    my ($self,$BR,$current_buildreq,$distroinfo,$db,$list,$flag_not_found_ref,$ignore,$input_drop)=@_;
    &__push($BR, map {&__check_dups($self,$current_buildreq,$distroinfo,$db,$_,$flag_not_found_ref,$ignore,$input_drop)} @$list);
}

sub __add_resolvable_br2 {
    my ($self,$BR,$IGNORE,$optsuffix,$mapper,$current_buildreq,$distroinfo,$db,$storage,$flag_not_found_ref,$global_ignore,$input_drop)=@_;
    my $key=$STORAGE_KEY{$db};
    die "internal error: no storage key for $db" unless $key;
    &__add_resolvable_br($self,$IGNORE,$optsuffix,$mapper,$current_buildreq,$distroinfo,$db,$storage->{$key.'_ignore'},undef,{},{});
    # no warnings about unresolved on ignores...
    my $local_ignore = {(%$global_ignore), map {&$mapper($_)=>1} @{$storage->{$key.'_ignore'}}};
    #print STDERR "local ignore: ",join(' ',@{$storage->{$key.'_ignore'}}),"\n";
    &__add_resolvable_br($self,$BR,$optsuffix,$mapper,$current_buildreq,$distroinfo,$db,$storage->{$key.'req'},$flag_not_found_ref,$local_ignore,$input_drop);
}

sub __add_resolvable_br {
    my ($self,$BR,$optsuffix,$mapper,$current_buildreq,$distroinfo,$db,$list,$flag_not_found_ref,$ignore,$input_drop)=@_;
    $mapper ||= sub { return $_[0]};
    if ($self->{-resolve_.$optsuffix}) {
	&__add_br($self,$BR,$current_buildreq,$distroinfo,$db,$list,$flag_not_found_ref,$ignore,$input_drop);
    } else {
	foreach my $name (@$list) {
	    my @data=&__check_dups($self,$current_buildreq,$distroinfo,$db,$name,$flag_not_found_ref,$ignore,$input_drop);
	    push @$BR, &$mapper($name) if (@data or $self->{-allow_unknown_.$optsuffix}) and
		not ($self->{-skip_implicit_.$optsuffix} and &__hash_have_any($current_buildreq,\@data));
	}
    }
}

sub __uniquify_storage {
    my ($storage)=@_;
    foreach (qw/libs hdrs path pkgconf girs perl/) {
	my @new_list=&__unique_list($storage->{$_.'_ignore'},@{$storage->{$_.'req'}});
	$storage->{$_}=\@new_list;
    }
}

sub __debug_storage {
    my ($storage,$inputname)=@_;
    &__debug_msg($inputname,'BR girs',$storage->{girsreq});
    &__debug_msg($inputname,'BR devel-libs',$storage->{libsreq});
    &__debug_msg($inputname,'BR headers',$storage->{hdrsreq});
    &__debug_msg($inputname,'BR path',$storage->{pathreq});
    &__debug_msg($inputname,'BR perl',$storage->{perlreq});
    &__debug_msg($inputname,'BR pkg-config',$storage->{pkgconfreq});
    &__debug_msg($inputname,'ignore girs',$storage->{girs_ignore});
    &__debug_msg($inputname,'ignore devel-libs',$storage->{libs_ignore});
    &__debug_msg($inputname,'ignore headers',$storage->{hdrs_ignore});
    &__debug_msg($inputname,'ignore path',$storage->{path_ignore});
    &__debug_msg($inputname,'ignore perl',$storage->{perl_ignore});
    &__debug_msg($inputname,'ignore pkg-config',$storage->{pkgconf_ignore});
}

sub __debug_msg {
    my ($file,$br_name,$list)=@_;
    return unless @$list;
    print STDERR "$file: $br_name(detected):",
    join(' ',map {ref($_) eq 'ARRAY' ? '[ '.join(' ',@$_).' ]' : $_} @$list),"\n";
}

# if $flag_not_found_ref is not supplied than no warning is issued.

sub __check_dups {
    my ($self,$current_buildreq,$distroinfo,$db,$name,$flag_not_found_ref,$ignore,$input_drop)=@_;
    confess "Oops! __check_dups: invalid ignore $ignore: shoud be hash" if $flag_not_found_ref and ref $ignore ne 'HASH';
    confess "Opps! __check_dups: invalid input_ignore $input_drop: shoud be hash" if $input_drop and ref $input_drop ne 'HASH';
    my @names=($name);
    @names=@$name if ref($name) eq 'ARRAY';
    my @data;
    foreach my $candidate (@names) {
	#print STDERR "SourceAnalyzer: checking: $candidate\n";
	next if $input_drop and $input_drop->{$candidate};
	my @data0=$distroinfo->translate($db,$candidate);
	return() if @names>1 and &__hash_have_any($current_buildreq,\@data0);
	push @data, @data0;
	last if @data;
    }
    if (@data>1) {
	return() if &__hash_have_any($current_buildreq,\@data);
	if (@names==1) {
	    if ($self->{-select_multiple_dependency}->{$names[0]}) {
		@data=$self->{-select_multiple_dependency}->{$names[0]};
	    } elsif ($opt_select_multiple_dependency{$names[0]}) {
		@data=($opt_select_multiple_dependency{$names[0]});
	    }
	}
	warn 'WARNING: SourceAnalyzer: multiple provides in '.$db.' for '.join(' ',@names).': '.join(' ',@data)."\n" if @data>1;
    } elsif (not @data and $flag_not_found_ref and &__names_are_not_ignored($ignore,$input_drop,@names)) {
	print STDERR "INFO: SourceAnalyzer: nothing in $db provides ".join(' ',@names).".\n";
	$$flag_not_found_ref=1;
    }
    return @data;
}

sub __names_are_not_ignored {
    my ($ignore,$drop,@names)=@_;
    return 1 if not $ignore;
    foreach my $name (@names) {
	return 1 if not $ignore->{$name} and not ($drop and $drop->{$name});
    }
    return 0;
}

sub __hash_have_any {
    my ($hash,$list)=@_;
    foreach my $key (@$list) {
	return 1 if $hash->{$key};
    }
    return 0;
}

my %LANG_FLAGS=qw/
AC_PROG_CC	c
AM_PROG_CC_STDC	c
AM_PROG_LIBTOOL	c
AC_PROG_CXX	cxx
AC_PROG_CXXCPP	cxx
AC_PROG_F77	fortran
AC_LANG_C	c
AC_LANG_CPLUSPLUS	cxx
AC_LANG_FORTRAN77	fortran
/;

my %LIB_FLAGS=qw/
AM_PATH_AALIB	aa
AM_PATH_ALLEGRO	alleg
AM_PATH_ALSA	asound
AM_PATH_ASPELL	aspell
AC_PATH_BLUETOOTH	bluetooth
AM_PATH_CHECK	check
AC_PATH_CLN	cln
AM_PATH_CPPUNIT	cppunit
AM_PATH_DVDNAV	dvdnav
AM_PATH_ESD	esd
AM_PATH_FLITE	flite
AM_PATH_FLTK	fltk
AM_PATH_CG	cg
AM_PATH_GINAC	ginac
AM_PATH_GMP	gmp
AM_PATH_GPG_ERROR	gpg-error
AM_PATH_GSL	gsl
AM_PATH_GTK	gtk
AM_PATH_GTK_2_0	gtk-x11-2.0
AM_PATH_GTK_3_0 gtk-3
AM_PATH_GLIB	glib
AM_PATH_GLIB_2_0	glib-2.0
AM_PATH_GLUT	glut
AM_PATH_GPGME	gpgme
AM_PATH_LAPACK	lapack
AM_PATH_LIB3DS	3ds
AM_PATH_LIBEXPAT	expat
AM_PATH_LIBFAME	fame
AM_PATH_LIBFLAC	flac
AM_PATH_LIBGCRYPT_CCRTP	gcrypt
AM_PATH_LIBGCRYPT	gcrypt
AM_PATH_LIBGNUTLS	gnutls
AM_PATH_LIBMIKMOD	mikmod
AM_PATH_LIBOTR		otr
AM_PATH_LIBPNG		png
AM_PATH_LIBPRELUDE	prelude
AM_PATH_LIBRAPI2	rapi
AM_PATH_LUA	lua
AC_PATH_MAGIC	magic
AM_PATH_MPFR	mpfr
AM_PATH_ODBC	odbc
AM_PATH_OPENGL	GL
AM_PATH_OPENOBEX	openobex
AM_PATH_PHYSFS	physfs
AM_PATH_PILOT_LINK	pisync
AM_PATH_PNG	png
AM_PATH_RADIUS_ENGINE	radius-engine
AM_PATH_READLINE	readline
AM_PATH_SC	mpqc
AM_PATH_SDL	SDL
AM_PATH_SDL_GFX	SDL_gfx
AM_PATH_SDL_IMAGE	SDL_image
AM_PATH_SDL_MIXER	SDL_mixer
AM_PATH_SDL_NET	SDL_net
AM_PATH_SDL_PANGO	SDL_pango
AM_PATH_SDL_SOUND	SDL_sound
AM_PATH_SDL_TTF	SDL_ttf
AM_PATH_STDAIR	stdair
AM_PATH_TALKFILTERS	talkfilters
AC_PATH_USB	usb-1.0
AM_PATH_WNN	wnn;crypt
AM_PATH_XFT	Xft
AM_PATH_XINE	xine
AM_PATH_XML2	xml2
AM_PATH_XMMS	xmms
AC_PATH_XTRA	ICE;SM
AST_IMLIB2_SUPPORT	Imlib2;freetype;Xext
DAP_CHECK_ZLIB	z
DECAF_CONFIGURE_APR	expat
/;

my %HEADER_FLAGS=qw!
AC_PATH_IRDA	lib/irda.h
AC_PATH_IRDA_LINUX	lib/irda.h
AM_PATH_SYSLOG	syslog.h
AC_PATH_X	X11/Xlib.h
!;

my %PKG_FLAGS=qw!
AM_PATH_GNAT	gcc-gnat
AM_PATH_HASHDIR ispell
AC_PATH_KDE	kdelibs-devel
AM_PATH_LIBGLADE	libglade-devel
AM_PATH_PKGCONFIG	pkg-config
AM_PATH_PYGTK	python-module-pygtk-devel
AM_PATH_PYTHON	python-devel
AM_PATH_PYTHON_VERSION	python-devel
AC_PATH_QT_1_3	libqt3-devel
AM_PATH_RUBY	libruby-devel
AM_PATH_SQLITE	sqlite3
GLIB_GSETTINGS	libgio-devel
!;

my %PATH_FLAGS=qw!
GTK_DOC_CHECK	gtkdocize
AM_GLIB_GNU_GETTEXT	glib-gettextize
AM_PATH_CCACHE	ccache
AM_PATH_PERL	perl
AM_PATH_SQLITE3	sqlite3
AM_PATH_WXCONFIG	wx-config
!;

my %IGNORE_FLAGS=map {$_=>1} qw!
AM_PATH_LISPDIR	
AC_PATH_WIN32	
AM_PATH_DIRECTX	
AM_PATH_LIBOLE2	
AM_PATH_ARTS	
AC_PATH_TOOL	
AC_PATH_GENERIC
!;

#GNOME_COMMON_INIT
## Docs are disabled for now
##GNOME_DOC_INIT
#GNOME_MAINTAINER_MODE_DEFINES
#GNOME_COMPILE_WARNINGS
#GNOME_DEBUG_CHECK

# TODO: implicit AC_TRY_LINK
#LIBS="-lutil"
#AC_TRY_LINK(, , [needutil=yes], [needutil=no])

# dep on GConf
#AM_GCONF_SOURCE_2

sub _autoconf_macro_handler {
    my ($name,$args,$ac_parser)=@_;
    my $storage=$ac_parser->{-storage};
    my $vars=$ac_parser->{-vardata};
    my $lang_flag=$LANG_FLAGS{$name};
    my $lib_flag=$LIB_FLAGS{$name};
    my $pkg_flag=$PKG_FLAGS{$name};
    my $path_flag=$PATH_FLAGS{$name};
    my $header_flag=$HEADER_FLAGS{$name};
    $storage->{lang}->{$lang_flag}=1 if $lang_flag;
    push @{$storage->{libsreq}}, split(';',$lib_flag) if $lib_flag;
    push @{$storage->{pkgsreq}}, split(';',$pkg_flag) if $pkg_flag;
    push @{$storage->{pathreq}}, split(';',$path_flag) if $path_flag;
    push @{$storage->{hdrsreq}}, &__add_h(split(';',$header_flag)) if $header_flag;
    warn "SourceAnalyzer: unknown macros $name. Please. report.\n" if $name=~/^A[CM]_PATH_(?!PROG)/ and not ($lang_flag or $path_flag or $lib_flag or $header_flag or $pkg_flag or $IGNORE_FLAGS{$name});

    if ($name eq 'AC_LANG') {
	my $lang=__strip($args->[0]);
	if ($lang eq 'C') {
	    $storage->{lang}->{'c'}=1;
	} elsif ($lang eq 'C++') {
	    $storage->{lang}->{'cxx'}=1;
	} elsif ($lang =~ '^Fortran') {
	    $storage->{lang}->{'fortran'}=1;
	} else {
	    warn "SourceAnalyzer bug: unknown language in AC_LANG($lang). Please, report.\n";
	}
    } elsif ($name eq 'AC_CHECK_FILE') {
	if ($args->[0] =~ /(docbook|titlepage).xsl/) {
	    push @{$storage->{pkgsreq}}, 'docbook-dtds', 'docbook-style-xsl';
	} else {
	    warn "SourceAnalyzer: AC_CHECK_FILE '$args->[0]'\n";
	}
    } elsif ($name eq 'AC_CHECK_LIB') {
	&__push($storage->{libsreq}, &__subst_vars($args->[0],$vars));
    } elsif ($name eq 'AC_SEARCH_LIBS') {
	&__push($storage->{libsreq}, [&__split(&__subst_vars($args->[1],$vars))]);
    } elsif ($name eq 'AC_CHECK_HEADER') {
	&__push($storage->{hdrsreq}, &__add_h(&__subst_vars($args->[0], $vars)));
    } elsif ($name eq 'AC_CHECK_HEADERS') {
	&__push($storage->{hdrsreq}, &__add_h(&__split(&__subst_vars($args->[0],$vars))));
    } elsif ($name eq 'AC_PATH_GENERIC') {
	&__push($storage->{pathreq}, map {$_.'-config'} &__subst_vars($args->[1],$vars));
    } elsif ($name eq 'AC_PATH_PROG') {
	my $path=$args->[1];
	if (defined $path) {
	    &__push($storage->{pathreq}, &__subst_vars($path,$vars));
	} else {
	    warn "SourceAnalyzer: strange AC_PATH_PROG ($args->[0]) - no second arg is given"
	}
    } elsif ($name eq 'AC_CHECK_PROG') {
	&__push($storage->{pathreq}, &__subst_vars($args->[1],$vars));
    } elsif ($name eq 'AC_CHECK_PROGS') {
	&__push($storage->{pathreq}, [&__split(&__subst_vars($args->[1],$vars))]);
    } elsif ($name eq 'JH_CHECK_XML_CATALOG') {
	if ($args->[0] =~ /DTD DocBook/) {
	    push @{$storage->{pkgsreq}}, 'docbook-dtds';
	} elsif ($args->[0] =~ /docbook.xsl/) {
	    push @{$storage->{pkgsreq}}, 'docbook-style-xsl';
	} else {
	    warn "SourceAnalyzer: unrecognized XML catalog '$args->[0]'\n";
	}
    } elsif ($name eq 'PKG_CHECK_MODULES') {
	my $pkgline=&__subst_vars($args->[1],$vars);
# note 3 arg
#PKG_CHECK_MODULES([ENCHANT], [enchant], [AC_DEFINE(HAVE_LIBENCHANT, , Define if we use enchant)], [])
# TODO: multiline (see assogiate)
#PKG_CHECK_MODULES([ASSOGIATE], [                                                
#        glib-2.0 >= 2.8.0
#        glibmm-2.4 >= 2.8.0
#        gtkmm-2.4 >= 2.6.0
#        libxml++-2.6
#        gnome-vfsmm-2.6
#])                                                                              
# TODO: variable (see xqf); note: it is inside if; so it still might be better to ignore
#        pkg_modules="gtk+-2.0 >= 2.0.0 gdk-pixbuf-xlib-2.0"
#        PKG_CHECK_MODULES(PACKAGE, [$pkg_modules])


	my @raw_pkgconf;
	foreach my $item (grep {$_} split (/\s+/, $pkgline)) {
	    if ($item=~/^.+\$([\w_][\w_\d]*)/) {
		if (not $vars->{$1}) {
		    push @raw_pkgconf, $item;
		} else {
		    #print STDERR "var: $1 - ",join(':',@{$vars->{$1}}),"\n";
		    foreach (@{$vars->{$1}}) {
			my $expanded_item=$item;
			$expanded_item=~ s/^.+\K\$([\w_][\w_\d]*)/$_/;
			push @raw_pkgconf, $expanded_item;
		    }
		}
	    } else {
		push @raw_pkgconf, $item;
	    }
	}
	while (@raw_pkgconf>0) {
	    my $in=shift @raw_pkgconf;
	    next if $in =~ /^\$/;
	    $in =~s/^['"]//;
	    $in =~s/['"]$//;
	    if ($in =~ /^[<>=!]/) {
		shift @raw_pkgconf;
		next;
	    }
	    push @{$storage->{pkgconfreq}}, $in;
	}
    }
}

sub __subst_vars {
    my ($in,$vars)=@_;
    my $line=&__strip($in);
    $line=~s/(?:^|\s)\K\$([\w_][\w_\d]*)/$vars->{$1} ? join(' ',@{$vars->{$1}}) : $1/eg;
    return $line;
}

sub __push {
    my ($ref,@in)=@_;
    map {Carp::cluck "undefined value" unless defined $_} @_;
    push @$ref, @in;
}

sub __strip {
    local $_=$_[0];
    Carp::cluck "undefined value" unless defined $_;
    s/\\\n//gs;
    s/^\s*//;
    s/\s*$//;
    return $_;
}

sub __split {
    my ($line)=@_;
    return grep {defined($_) and $_ ne ''} split (/\s+/, __strip($line));
}

sub __add_h {
    # not used...
    #return map {/\.h$/ ? $_ : $_.'.h'} @_;
    return @_;
}

sub __add_file_to_flag {
    my ($file,$hash)=@_;
    return unless -e $file;
    open my $fh, "<", $file or die $!;
    local $_;
    while (<$fh>) {
	chomp;
	s/^\s*//;
	s/\s*$//;
	$hash->{$_}=1;
    }
    close $fh;
}

sub _automake_var_handler {
    my ($name,$args,$am_parser)=@_;
    my $storage=$am_parser->{-storage};
    if ($am_parser->{-verbose}>3) {
	print "_automake_var_handler: $name=$args\n";
    }
    if ($name=~/_(?:LIBS|LDFLAGS|LDADD|LIBADD)$/) {
	foreach my $word (shellwords($args)) {
	    &__push($storage->{libsreq}, $word) if $word=~s/^-l(?=\w)//;
	}
    } elsif ($name=~/_gir_INCLUDES$/) {
	foreach my $word (shellwords($args)) {
	    if ($word eq 'Gtk-${GTK_API_VERSION}' or $word eq 'Gtk-$(GTK_API_VERSION)') {
		&__push($storage->{girsreq}, 'Gtk-2.0');
		&__push($storage->{girsreq}, 'Gtk-3.0');
	    } else {
		&__push($storage->{girsreq}, $word);
	    }
	}
    }
}

sub _qmake_var_handler {
    my ($name,$args,$parser)=@_;
    my $storage=$parser->{-storage};
    print "_qmake_var_handler: $name=$args\n" if $parser->{-verbose}>3;
    if ($name eq 'LIBS') {
     	foreach my $word (shellwords($args)) {
     	    &__push($storage->{libsreq}, $word) if $word=~s/^-l(?=\w)//;
     	    print STDERR "qmake project: TODO: need lib${word}.so\n" if $word=~s/^-l(?=\w)// and $parser->{-verbose};
     	}
    } elsif ($name eq 'PKGCONFIG') {
     	foreach my $word (shellwords($args)) {
     	    &__push($storage->{pkgconfreq}, $word) if $word=~/^\w/;
     	}
    } elsif ($name eq 'TEMPLATE') {
	# if 
        #TEMPLATE = lib
	# then put target to -ignore section (internal lib)?
	# or just use flag 'ban self'
	$parser->{'__TEMPLATE'}=$args;
    } elsif ($name eq 'TARGET') {
	$parser->{'__TARGET'}=$args;
    } elsif ($name eq 'QT') {
	# TODO proper qt3,4 selection
	$storage->{lang}->{'cxx'}=1;
    }
}

sub _cmake_path_handler {
    my ($name,$parser)=@_;
    my $storage=$parser->{-storage};
    print "_cmake_path_handler: $name\n" if $parser->{-verbose}>3;
    &__push($storage->{pathreq}, $name);
}

sub _cmake_postfile_handler {
}

sub _qmake_postfile_handler {
    my ($parser)=@_;
    my $template=$parser->{'__TEMPLATE'};
    my $target=$parser->{'__TARGET'};
    if ($target and $template and $template eq 'lib') {
	&__push($parser->{-storage}->{libs_ignore}, $target);
	print STDERR "I: ignored internal library $target\n" if $parser->{-verbose}>1;
    }
}

sub _cpp_header_handler {
    my ($name,$parser)=@_;
    my $storage=$parser->{-storage};
    print "_cpp_header_handler: $name\n" if $parser->{-verbose}>3;
    &__push($storage->{hdrsreq}, &__add_h($name));
}

package __::SourceAnalyzer::Collect;

use strict;
use warnings;

sub new {
    my $class = shift;
    my $self = {
	libsreq=>[],
	hdrsreq=>[],
	pathreq=>[],
	perlreq=>[],
	pkgsreq=>[],
	girsreq=>[],
	vapireq=>[],
	pkgconfreq=>[],
	libs_ignore=>[],
	hdrs_ignore=>[],
	path_ignore=>[],
	perl_ignore=>[],
	pkgs_ignore=>[],
	girs_ignore=>[],
	vapi_ignore=>[],
	pkgconf_ignore=>[],
	lang=>{},
	@_
    };
    bless $self, $class;
}

package SourceAnalyzer::Language;

use strict;
use warnings;
use Cwd;
use File::Basename;
our $VERSION=0.001;

sub parse_file {
    my ($self,$file)=@_;
    my $content = &__load_file($file);
    return $self->parse_string($content);
}

sub __load_file {
    my ($file)=@_;
    open my $fh, "<", $file or do {
	print STDERR "Internal error: __load_file: when opening $file:\nin curdir=",Cwd::getcwd()," got error: $!\n";
	my $path=$file;
	while (not -e $path) {
	    print STDERR "path that not found: [$path]\n";
	    $path=dirname($path);
	}
	print STDERR "top path that found: $path\n\ls:\n";
	system('ls',$path);
	Carp::confess();
    };
    local $/; # enable localized slurp mode
    my $content = <$fh>;
    close $fh;
    return $content;
}

sub _raw_collect_handler {
    my $parser=shift;
    my $type=shift;
    my $storage=$parser->{-storage};
    #print STDERR "_raw_collect_handler: P:$parser S:$storage T:$type ".join(' ',$@)."\n";
    if ($parser->{-verbose}>3) {
	print "_raw_collect_handler: $type ".join(' ',$@)."\n";
    }
    &RPM::Source::Tools::SourceAnalyzer::__push($storage->{$type}, @_);
}


package SourceAnalyzer::Language::Autoconf;

use strict;
use warnings;
our $VERSION=0.001;
our @ISA=qw/SourceAnalyzer::Language/;

# for copypaste from Text/ParseWords.pm; seems to be support for perlish syntax not useful for us.
my $PERL_SINGLE_QUOTE;
our $verbose=0;

sub new {
    my $class = shift;
    my $self = {
	-vars => {},
	-verbose => 0,
	-varhook => \&_var_handler,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $self->{-vardata}={};
    my $pure_data='';
    while (length($data)) {
	&__purge_dnl(\$pure_data,\$data);
    }
    while (length($pure_data)) {
	&_read_any($self,\$pure_data);
    }
}

# C++ features
#CXXAC_TYPE_BOOL
#CXXAC_BAD_ALLOC
#CXXAC_HAVE_EXPLICIT
#CXXAC_HAVE_MUTABLE
#CXXAC_HAVE_ARRAY_NEW
#CXXAC_BAD_DELETE
#CXXAC_DEFAULT_EXCEPTION
#CXXAC_HAVE_TYPENAME
#CXXAC_NO_GUIDE_DECL
#CXXAC_HEADER_ALL

my %KNOWN_MACRO=map {$_=>1} qw!
DAP_CHECK_ZLIB
DECAF_CONFIGURE_APR
DECAF_CONFIGURE_SSL
GNOME_COMMON_INIT
GNOME_COMPILE_WARNINGS
GNOME_DEBUG_CHECK
GNOME_DOC_INIT
PKG_CHECK_MODULES
PKG_PROG_PKG_CONFIG
GLIB_GSETTINGS
GTK_DOC_CHECK
AST_IMLIB2_SUPPORT
JH_CHECK_XML_CATALOG
!;

sub __purge_dnl {
    my ($outref,$inref)=@_;
    my $read='';
    if ($$inref=~s/^([^\n]*\n)//) {
	$read=$1;
    } else {
	$read=$$inref;
	$$inref='';
    }
    $read=~s/\b(?:[mM]4_)?[Dd][Nn][Ll].*$//s;
    $$outref .= $read;
}

sub _read_any {
    my ($self,$ref)=@_;
    my $read=&_read_non_token_non_end_or_comments($ref);
    return $read if $read ne '';
    $read=$1 if $$ref=~s/^([,\)])//;
    return $read if $read ne '';
    my $token=&_read_token($ref);
    warn "Oops! no token" if $token eq '';
    $read.=$token;
    if ($token eq 'dnl') {
	$$ref=~s/^([^\n]*\n)//s;
	return $read.$1;
    } elsif ($token eq 'm4_pattern_forbid' or $token eq 'm4_pattern_allow' or $token eq 'pattern_forbid' or $token eq 'pattern_allow') {
	my $args=[];
	$read.=&_read_args($self,$ref,$args) if $$ref=~/^\(/;
    } elsif ($token eq 'ifdef' or $token eq 'm4_ifdef' or $token eq 'ifndef' or $token eq 'm4_ifndef') {
	# read name of macro tested, it can trigger events
	my ($unquoted, $raw)=&_read_shellword('\s+', $ref);
	$read .= $raw;

    } elsif ($KNOWN_MACRO{$token} or $token=~/^(?:AC_|AM_|CXXAC_)/) {
	my $args=[];
	$read.=&_read_args($self,$ref,$args) if $$ref=~/^\(/;
	$self->add_macro($token,$args);
    } elsif ($$ref=~s/^=//) {
	my ($unquoted, $raw)=&_read_shellword('\s+', $ref);
	$self->add_var($token,&__kill_dnl($unquoted));
	$read .= '='.$raw;
    }
    return $read;
}

sub add_macro {
    my ($self,$name,$args)=@_;
    print STDERR "MACRO: $name ARGS:",join(',',@$args),"\n" if $verbose>1;
    my $hook=$self->{-macrohook};
    &$hook($name,$args,$self) if $hook;
}

sub add_var {
    my ($self,$name,$val)=@_;
    print STDERR "VAR: add $name=$val\n" if $verbose>1;
    my $hook=$self->{-varhook};
    &$hook($name,$val,$self) if $hook;
    return $val;
}

sub _var_handler {
    my ($name,$val,$self)=@_;
    my $VAR=$self->{-vardata};
    $val=$1 if $val and $val=~/^\[(.*)\]$/;
    my $ref=$VAR->{$name};
    unless ($ref) {
	$ref=[];
	$VAR->{$name}=$ref;
    }
    push @$ref, $val;
}

sub _read_token {
    my ($ref)=@_;
    if ($$ref=~s/^([\w_][\w_\d]*)//s) {
	print STDERR "read token: $1\n" if $verbose>2;
	return $1;
    } else {
	return '' if $$ref eq '';
	die "Oops! read_token - no token :( ->".substr($$ref,0,1);
    }
}

sub _read_non_token_non_end_or_comments {
    my ($ref)=@_;
    my $read='';
    while (length($$ref) and $$ref !~/^[\w_\),]/) {
	if ($$ref=~s/^(\s+)//s) {
	    $read.=$1;
	} elsif ($$ref=~s/^(#[^\n]*)\n//s) {
	    print STDERR "drop comment: $1\n" if $verbose>2;
	    $read.=$1;
	} else {
	    $$ref=~s/^(.)//s;
	    $read.=$1;
	}
    }
    return $read;
}

sub _read_args {
    my ($self,$ref,$arg)=@_;
    my $read='';
    my $cur_arg='';
    if ($$ref=~s/^(\()//s) {
	$read.=$1;
    }
    while (length($$ref) and $$ref !~/^\)/) {
	if ($$ref=~s/^(\s+)//s) {
	    $read.=$1;
	    $cur_arg.=$1;
	# TODO: properly read m4 quoted string
	# TODO: load shell var assignments in m4 quoted string
	} elsif ($$ref=~s/^(\[([^]]*)\])//s) {
	    $read.=$1;
	    $cur_arg.=&__kill_dnl($2);
	} elsif ($$ref=~s/^(,)//s) {
	    $read.=$1;
	    push @$arg, $cur_arg;
	    $cur_arg='';
	} elsif ($$ref=~s/^(dnl(?:\n|\s[^\n]*\n))//s) {
	    $read.=$1;
	} else {
	    $cur_arg.=&_read_any($self,$ref);
	}
    }
    if ($$ref=~s/^(\))//s) {
	$read.=$1;
    }
    push @$arg, $cur_arg;
    return &__kill_dnl($read);
}

sub __kill_dnl {
    my $in=shift;
    $in=~s/^dnl\n/\n/smg;
    $in=~s/^dnl\s[^\n]*\n/\n/smg;
    $in=~s/(?<=\W)dnl\n/\n/sg;
    $in=~s/(?<=\W)dnl\s[^\n]*\n/\n/sg;
    return $in;
}

# regexp is copypasted from Text/ParseWords.pm
sub _read_shellword {
    my($delimiter, $lineref) = @_;
    my($word);

    no warnings 'uninitialized';	# we will be testing undef strings

    return '' if not length($$lineref);
        # This pattern is optimised to be stack conservative on older perls.
        # Do not refactor without being careful and testing it on very long strings.
        # See Perl bug #42980 for an example of a stack busting input.
    $$lineref =~ s/^
                    (?: 
                        # double quoted string
                        (")                             # $quote
                        ((?>[^\\"]*(?:\\.[^\\"]*)*))"   # $quoted 
		    |	# --OR--
                        # singe quoted string
                        (')                             # $quote
                        ((?>[^\\']*(?:\\.[^\\']*)*))'   # $quoted
                    |   # --OR--
                        # unquoted string
		        (                               # $unquoted 
                            (?:\\.|[^\\"'])*?           
                        )		
                        # followed by
		        (                               # $delim
                            \Z(?!\n)                    # EOL
                        |   # --OR--
                            (?-x:$delimiter)            # delimiter
                        |   # --OR--                    
                            (?!^)(?=["'])               # a quote
                        )  
		    )//xs or return;		# extended layout                  
        my ($quote, $quoted, $unquoted, $delim) = (($1 ? ($1,$2) : ($3,$4)), $5, $6);


	return '' unless( defined($quote) || length($unquoted) || length($delim));

    my $raw = "$quote$quoted$quote$delim";

    $unquoted =~ s/\\(.)/$1/sg;
    if (defined $quote) {
	$quoted =~ s/\\(.)/$1/sg if ($quote eq '"');
	$quoted =~ s/\\([\\'])/$1/g if ( $PERL_SINGLE_QUOTE && $quote eq "'");
    }
    my $processed = defined $quote ? $quoted : $unquoted;
    return ($processed, $raw);
}

package SourceAnalyzer::Language::AutoMake;

use strict;
use warnings;

our @ISA=qw/SourceAnalyzer::Language/;
our $verbose=0;
our $VERSION=0.001;

sub new {
    my $class = shift;
    my $self = {
	-verbose => 0,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $data=~s/\\\n//gs;
    foreach my $line (split("\n",$data)) {
	&_parse_line($self,$line);
    }
    my $hook=$self->{-postfilehook};
    &$hook($self) if $hook;
}

sub _parse_line {
    my ($self,$line)=@_;
    # Makefile is line-censitive
    #$line=~s/^\s*//;
    $line=~s/\s*$//;
    #print STDERR "LINE=$line\n";
    if ($line=~s/^AM_VALAFLAGS\s*=\s*//) {
	#print STDERR "AM_VALAFLAGS=$line\n";
	my @tokens=split(/\s+/,$line);
	while (my $token = shift @tokens) {
	    #print STDERR "AM_VALAFLAGS: token=$token\n";
	    if ($token=~/^--pkg=(.*)$/) {
		#print STDERR "  pkg=$1\n";
		$self->_raw_collect_handler('vapireq',$1);
	    } elsif ($token eq '--pkg') {
		$token = shift @tokens;
		#print STDERR "AM_VALAFLAGS: vapi=$token\n";
		$self->_raw_collect_handler('vapireq',$token) if $token;
	    } elsif ($token=~/^--gir=(.*)$/) {
		#print STDERR "  gir=$1\n";
		$self->_raw_collect_handler('girsreq',$1);
	    } elsif ($token eq '--gir') {
		$token = shift @tokens;
		#print STDERR "AM_VALAFLAGS: gir=$token\n";
		$self->_raw_collect_handler('girsreq',$token) if $token;
	    }
	}
    } elsif ($line=~/^([\w_][\w\d_]*)\s*\+?=\s*(.*)$/) {
	$self->add_var($1,$2);
    } elsif ($line=~/^([\w_][\w\d_]*):\s*(.*)$/) {
	# target
    } elsif ($line=~s/^.*\s\$\((?:G_IR|INTROSPECTION)_SCANNER\)//) {
	# $(G_IR_SCANNER) $(INTROSPECTION_SCANNER)
	# cmds
	#print STDERR "girline=$line\n";
	foreach my $token (split(/\s+/,$line)) {
	    #print STDERR "  token=$token\n";
	    if ($token=~/^--include=(.*)$/) {
		#print STDERR "  include=$1\n";
		if ($1 eq 'Gtk-${GTK_API_VERSION}' or $1 eq 'Gtk-$(GTK_API_VERSION)') {
		    $self->_raw_collect_handler('girsreq','Gtk-2.0');
		    $self->_raw_collect_handler('girsreq','Gtk-3.0');
		} else {
		    $self->_raw_collect_handler('girsreq',$1);
		}
	    } elsif ($token=~/^--pkg=(.*)$/) {
		#print STDERR "  pkg=$1\n";
		if ($1 eq 'gtk+-${GTK_API_VERSION}' or $1 eq 'gtk+-$(GTK_API_VERSION)') {
		    $self->_raw_collect_handler('pkgconfreq','gtk+-2.0');
		    $self->_raw_collect_handler('pkgconfreq','gtk+-3.0');
		} else {
		    $self->_raw_collect_handler('pkgconfreq',$1);
		}
	    }
	}
    }
}

sub add_var {
    my ($self,$name,$val)=@_;
    print STDERR "VAR: add $name=$val\n" if $verbose>1;
    my $hook=$self->{-varhook};
    &$hook($name,$val,$self) if $hook;
    return $val;
}

package SourceAnalyzer::Language::QMake;

use strict;
use warnings;

our @ISA=qw/SourceAnalyzer::Language/;
our $verbose=0;
our $VERSION=0.001;

sub new {
    my $class = shift;
    my $self = {
	-verbose => 0,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $data=~s/\\\n//gs;
    foreach my $line (split("\n",$data)) {
	&_parse_line($self,$line);
    }
    my $hook=$self->{-postfilehook};
    &$hook($self) if $hook;
}

sub _parse_line {
    my ($self,$line)=@_;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    if ($line=~/([\w_][\w\d_]*)\s*\+?=\s*(.*)$/) {
	$self->add_var($1,$2);
    }
}

sub add_var {
    my ($self,$name,$val)=@_;
    print STDERR "VAR: add $name=$val\n" if $verbose>1;
    my $hook=$self->{-varhook};
    &$hook($name,$val,$self) if $hook;
    return $val;
}

package SourceAnalyzer::Language::CPP_header;

use strict;
use warnings;

our @ISA=qw/SourceAnalyzer::Language/;
our $verbose=0;
our $VERSION=0.001;

sub new {
    my $class = shift;
    my $self = {
	-verbose => 0,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    #$data=~s/\\\n//gs;
    foreach my $line (split("\n",$data)) {
	&_parse_line($self,$line);
    }
}

sub _parse_line {
    my ($self,$line)=@_;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    if ($line=~/^\s*#\s*include\s*<([^>]+)>$/) {
	$self->add_header($1);
    }
}

sub add_header {
    my ($self,$name)=@_;
    print STDERR "HEADER: add $name\n" if $verbose>1;
    my $hook=$self->{-headerhook};
    &$hook($name,$self) if $hook;
    return;
}

package SourceAnalyzer::Language::CMake;

use strict;
use warnings;

our @ISA=qw/SourceAnalyzer::Language/;
our $verbose=0;
our $VERSION=0.001;

sub new {
    my $class = shift;
    my $self = {
	-verbose => 0,
	@_
    };
    bless $self, $class;
}

sub parse_string {
    my ($self,$data)=@_;
    $data=~s/\\\n//gs;
    foreach my $line (split("\n",$data)) {
	&_parse_line($self,$line);
    }
    my $hook=$self->{-postfilehook};
    &$hook($self) if $hook;
}

sub _parse_line {
    my ($self,$line)=@_;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    #FIND_PROGRAM(FLEX_EXECUTABLE NAMES flex PATHS /usr/bin DOC "Flex executable")
    if ($line=~/FIND_PROGRAM\(\w+\s+NAMES\s+(\w+)\s+PATHS\s+/) {
	$self->add_path($1);
    }
}

sub add_path {
    my ($self,$name)=@_;
    print STDERR "PATH: add $name\n" if $verbose>1;
    my $hook=$self->{-pathhook};
    &$hook($name,$self) if $hook;
    return $name;
}


1;
