package RPM::Source::Transform;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use RPM::Source::Editor;
use RPM::Source::Abstract::PriorityCollection;
use RPM::Source::TransformContainer;
use RPM::Source::Shared::Resource::Debug;
use RPM::Source::Transformation::DefaultPlayer;

our $VERSION=0.014;

#print STDERR "DEBUG:SUPPORT_CLASSES:\n\t",join("\n\t", @RPM::Source::TransformContainer::SUPPORT_CLASSES),"\n";
#print STDERR "DEBUG:RESOURCE_CLASSES:\n\t",join("\n\t", @RPM::Source::TransformContainer::RESOURCE_CLASSES),"\n";
#print STDERR "DEBUG:FACTORY_CLASSES:\n\t",join("\n\t", @RPM::Source::TransformContainer::FACTORY_CLASSES),"\n";
#print STDERR "DEBUG:PLAYER=$RPM::Source::TransformContainer::PLAYER\n";
#print STDERR "DEBUG:INPUT_HANDLER=$RPM::Source::TransformContainer::INPUT_HANDLER\n";

my $help;

# TODO: make them $opt{}'ed and documented in help
our ($opt_tmpdir,$opt_safe_mode);
#### used, like in srpmconvert-fc or jppconvert
our $cleanup=1;
our $verbose=1;

our @LONGOPT=(
    'cleanup!'	=> \$cleanup,
    'safe-mode'	=> \$opt_safe_mode,
    'tmpdir=s'	=> \$opt_tmpdir,
);

our @PROTECTED_OPTIONS=(
    'help'  => \$help,
    'v|verbose+'  => \$verbose,
    'q|quiet'   => sub { $verbose = 0 },
);
our @EXT_OPTS;

sub usage {
    my ($self)=@_;
    print "General options:
	-h, --help
	-v, --verbose (can be specified multiple times)
	-q, --quiet
";
    &longopt_usage();
}

sub longopt_usage {
    print "
Processing options:
  SRPM packing/unpacking options:
	--safe-mode restrain of doing unsafe operations such as spawning rpm
	--tmpdir /path/ - alternative tmpdir if default tmpdir is not big enough.
	--no-cleanup (do not remove tmp files; useful for debugging)
";
    RPM::Source::TransformContainer->longopt_usage;
}

sub new {
    my $class = shift;
    my $self = {
	'-specargs' => [],
	@_
    };
    bless $self, $class;
    my $not_embedded=$RPM::Source::TransformContainer::INPUT_HANDLER && ! $self->{-embedded};
    if ($not_embedded) {
	$self->get_options();
    }
    $self->process_options();
    $self->verify_options();
    $self->{-resourceobj}=RPM::Source::Shared::ResourceContainer->new();
    $self->_init_transformations();
    if ($not_embedded) {
	$RPM::Source::TransformContainer::INPUT_HANDLER->process_args();
	if (not $RPM::Source::TransformContainer::INPUT_HANDLER->verify_args()) {
	    print STDERR "ERROR: invalid or missing input arguments.\n";
	    exit(4);
	}
    }
    return $self;
}


sub get_longopt {
    my ($self)=@_;
#    print STDERR "DEBUG:SUPPORT_CLASSES:\n\t",join("\n\t", @RPM::Source::TransformContainer::SUPPORT_CLASSES),"\n";
#    print STDERR "DEBUG:ParentProvider=$RPM::Source::TransformContainer::ParentProvider\n";
    return @LONGOPT, RPM::Source::TransformContainer->get_longopt;
}

# can be used both as class and object method (for Embedded usage)
# in the past was used as class method in octave-package-builder
sub get_options {
    my ($self)=@_;
    my $result = GetOptions ($self->get_longopt,@PROTECTED_OPTIONS);
    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }
}

sub process_options {
    RPM::Source::TransformContainer->process_options;
}

sub verify_options {}

sub mk_opt {
#    my ($self)=@_;
    ### TODO: use some fixed keys hash here
    my %opt=(
#	- => $opt_,
	);
    map {$_->register_messages(\%opt)} @RPM::Source::TransformContainer::SUPPORT_CLASSES, @RPM::Source::TransformContainer::FACTORY_CLASSES; # not in @RESOURCE_CLASSES
    return \%opt;
}

sub _init_transformations {
    my ($self)=@_;
    my $resourceobj=$self->{-resourceobj};
    map {$_->init_factory($resourceobj)} @RPM::Source::TransformContainer::FACTORY_CLASSES;
}

sub transform_once {
    my ($self,$spec,$parent,$opt)=@_;

    my $resourceobj=$self->{-resourceobj};
    map {$_->prepare_resource($spec,$parent,$opt,$resourceobj)} @RPM::Source::TransformContainer::RESOURCE_CLASSES;

    my $actions=RPM::Source::Abstract::PriorityCollection->new();
    $actions->add(map {$_->transformations($opt,$resourceobj)} @RPM::Source::TransformContainer::FACTORY_CLASSES);

    if ($RPM::Source::TransformContainer::PLAYER->apply_collection($spec,$parent,$opt,$resourceobj,$actions) == RPM::Source::Transformation::Factory::TRANSFORMATION_FAILURE) {
	goto CLEANUP;
    }

    my $write_rpm_ok;
    if ($RPM::Source::TransformContainer::WRITER) {
	if ($opt_safe_mode and exists $opt->{-write_hasher_tar} and not $opt->{-write_hasher_tar}) {
	    print STDERR "WARNING: writing hasher tar due to the safe mode.\n";
	    $opt->{-write_hasher_tar}=1;
	}
	$write_rpm_ok=$RPM::Source::TransformContainer::WRITER->write($spec,$parent,$opt,$resourceobj);
    }
  CLEANUP:
    $parent->cleanup() if $parent;
    $spec->cleanup() if $write_rpm_ok || $cleanup;
    
}

sub get_spec {
    my ($self,$inref)=@_;
    my $spec;
    return $spec unless $inref;
    my @tmpdir_opts;
    if (not $cleanup and $inref->[0] eq 'SOURCERPM') {
    # all those tmp_dirname hacks are just to have the RPM::Source::Editor's guts at hand on crash.
	my $tmp_dirname=basename($inref->[1]);
	$tmp_dirname=~s!.src.rpm!!;
	$tmp_dirname=cwd()."/$tmp_dirname.$$";
	push @tmpdir_opts, 	TMPDIR => $tmp_dirname;
    }
    $spec=RPM::Source::Editor->new(
	VERBOSE=> $verbose,
	CLEANUP=> $cleanup,
	-tmpdir => $opt_tmpdir,
	-use_rpmbuild_be => $opt_safe_mode ? 0: 1,
	-use_shell_eval => $opt_safe_mode ? 0: 1,
	@tmpdir_opts,
	@{$self->{'-specargs'}},
	@$inref,
	);
    die "FATAL: failed to load the file ".$inref->[1].".\n" unless $spec;
    return $spec;
}

sub get_parent {
    my ($self, $parentref)=@_;
    my $parent;
    if ($parentref and defined $parentref->[1]) {
	my $parentfile=$parentref->[1];
	die "FATAL: file does not exist: $parentfile.\n" unless -e $parentfile;
	$parent=RPM::Source::Editor->new(
	    VERBOSE=> $verbose,
	    -tmpdir => $opt_tmpdir,
	    -use_rpmbuild_be => $opt_safe_mode ? 0: 1,
	    @$parentref,
	    );
	die "FATAL: failed to load the parent $parentfile.\n" unless $parent;
    }
    return $parent;
}

1;
