package RPM::Source::Transformation::DiffWritePlayer;

use strict;
use warnings;
use Carp;
use Cwd;
use File::Basename;
use File::Path;
use File::Temp qw/tempfile tempdir/;

use RPM::Source::Transformation::Component;
use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Component/;

push @RPM::Source::TransformContainer::SUPPORT_CLASSES, __PACKAGE__;
$RPM::Source::TransformContainer::PLAYER=__PACKAGE__;

my ($opt_diff_by_transformation_dir, $opt_diff_digest_mode);
our $opt_diff_by_srpm_dir='.';
our $opt_diff_author_name='NMU script';
our $opt_diff_author_email='srpmtools';
our $opt_diff_email_subject='Detailed Transformation log.';
my $opt_diff_write_unchanged=0;
my $opt_diff_raise_release=1;

our $verbose=0;

sub longopt_usage {
    print "
  WriteSpecDiff options:
	--diff-author-name <name>	Default is $opt_diff_author_name.
	--diff-author-email <email>	Default is $opt_diff_author_email.
	--diff-email-subject <text>	Default is $opt_diff_email_subject.
	--diff-write-digest	Enable digest mode.
	--diff-by-srpm-dir <dir> Diff output dir. Default is $opt_diff_by_srpm_dir.
	--diff-by-transformation-dir <dir> Diff per transformation output dir.
	--diff-write-unchanged		Write diff even if empty.
	--no-diff-raise-release		Do not raise release.
";
}

our @LONGOPT=(
    'diff-author-name=s'   => \$opt_diff_author_name,
    'diff-author-email=s'   => \$opt_diff_author_email,
    'diff-email-subject=s'   => \$opt_diff_email_subject,
    'diff-write-digest'   => \$opt_diff_digest_mode,
    'diff-by-srpm-dir=s'   => \$opt_diff_by_srpm_dir,
    'diff-by-transformation-dir'   => \$opt_diff_by_transformation_dir,
    'diff-raise-release!'   => \$opt_diff_raise_release,
    'diff-write-unchanged!'   => \$opt_diff_write_unchanged,
    );

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-diff_author_name => $opt_diff_author_name,
	-diff_author_email => $opt_diff_author_email,
	-diff_email_subject => $opt_diff_email_subject,
	-diff_by_srpm_dir => $opt_diff_by_srpm_dir,
	-diff_by_transformation_dir => $opt_diff_by_transformation_dir,
	-diff_digest_mode => $opt_diff_digest_mode,
	-diff_raise_release => $opt_diff_raise_release,
	-diff_write_unchanged => $opt_diff_write_unchanged,
	-diff_name => undef,
	-diff_specname => undef,
	});
}

# ------------------------

sub apply_collection {
    my ($class,$spec,$parent,$opt,$resourceobj,$actions)=@_;

    my @transformations=$actions->get;
    my $raise_release=pop @transformations;
    die "Oops! internal: not a raise release $raise_release->{-name}" if $raise_release->{-name} ne 'raise release';

    $opt->{-diff_digest_mode}=1 unless defined $opt->{-diff_digest_mode};
    $opt->{-diff_specname} ||= basename($spec->{SPECFILE});

    unless ($opt->{-diff_name}) {
	my $sourceid=$opt->{-diff_specname};
	$sourceid=~s/\.spec$//;
	if ($spec->{SOURCERPM}) {
	    $sourceid=basename($spec->{SOURCERPM});
	    $sourceid=~s/\.src\.rpm$//;
	}
	$opt->{-diff_name}=$sourceid;
    }
    my $diffname=$opt->{-diff_name};

    -d $opt->{-diff_by_srpm_dir} || mkpath([$opt->{-diff_by_srpm_dir}]);
    $class->__mk_TMPDIR();

    my $tmpdir=$class->__get_TMPDIR;
    my $oldspec="$tmpdir/a/".$opt->{-diff_specname};
    my $newspec="$tmpdir/b/".$opt->{-diff_specname};
    my $digestpatchfile="$opt->{-diff_by_srpm_dir}/$diffname-digest.diff";
    my $solidpatchfile="$opt->{-diff_by_srpm_dir}/$diffname.diff";
    
    my $timestamp=`LANG=C date`;
    chomp $timestamp;
    my $mailheader=qq|From $opt->{-diff_author_email} $timestamp
From: "$opt->{-diff_author_name}" <$opt->{-diff_author_email}>
|;
    if ($opt->{-diff_digest_mode} || $opt->{-diff_by_transformation_dir}) {
	mkpath([dirname($oldspec)]);
	&write_file($oldspec, &load_file($spec->{SPECFILE}));
    }
    my $firststate=$spec->get_spec();
    my $beforestate=$firststate;
    my $nextstate;

    foreach my $transformation (@transformations) {
	print STDERR "DiffWritePlayer: applying transformation ",$transformation->{-name},"\n" if $verbose;
	my $exit_code=$transformation->apply($spec,$parent,$opt,$resourceobj);
	if ($exit_code == RPM::Source::Transformation::Factory::TRANSFORMATION_FAILURE) {
	    #return RPM::Source::Transformation::Factory::TRANSFORMATION_FAILURE;
	} elsif (not defined $exit_code or $exit_code != RPM::Source::Transformation::Factory::TRANSFORMATION_OK) {
	    warn "transformation $transformation->{-name} returned invalid exit code\n";
	}
	print STDERR "DiffWritePlayer: applied transformation ",$transformation->{-name},"\n" if $verbose>1;

	if ($opt->{-diff_digest_mode} || $opt->{-diff_by_transformation_dir}) {
	    $nextstate=$spec->get_spec();
	    if ($beforestate ne $nextstate) {
		mkpath([dirname($newspec)]);
		&write_file($newspec, $spec->get_spec());
		my $name=$transformation->{-name};
		warn "transformation has no name" unless defined $name;
		my $message=$transformation->{-description};
		$message||='(message skipped)';
		my $subject=$transformation->{-subject};
		$subject||="$name transformation";
		my $mailboxheader=qq|$mailheader
Subject: $subject

$message
|;
		&write_file($digestpatchfile, -e $digestpatchfile ? "\n".$mailboxheader : $mailboxheader, '>>');
		$class->__diff($oldspec,$newspec,'>>',$digestpatchfile);
		$class->__diff($oldspec,$newspec,'>>',"$opt->{-diff_by_transformation_dir}/$name.diff") if $opt->{-diff_by_transformation_dir};
		system 'mv', '-f', $newspec, $oldspec;
		$beforestate=$nextstate;
	    }
	}
    }

    $nextstate=$spec->get_spec();
    if ($opt_diff_raise_release and ($firststate ne $nextstate or $opt->{-diff_write_unchanged})) {
	print STDERR "DiffWritePlayer: preparing raise_release transformation\n" if $verbose>1;
	$raise_release->apply($spec,$parent,$opt,$resourceobj);
	mkpath([dirname($oldspec)]);
	mkpath([dirname($newspec)]);
	&write_file($oldspec, &load_file($spec->{SPECFILE}));
	&write_file($newspec, $spec->get_spec());
	my $mailboxheader=qq|$mailheader
Subject: $opt->{-diff_email_subject}

|.$spec->get_default_changelog();

	&write_file($solidpatchfile, $mailboxheader, '>');
	$class->__diff($oldspec,$newspec,'>>',$solidpatchfile);
	unlink $oldspec;
	unlink $newspec;
    }
    &__del_TMPDIR();
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

END {
    &__del_TMPDIR();
}

sub __diff {
    my ($class,$oldspec,$newspec,$direction, $outfile)=@_;
    -e $oldspec && -e $newspec || warn "__diff: $oldspec or $newspec does not exists!";
    my $relativeold=$oldspec;
    my $relativenew=$newspec;
    my $tmpdir=$class->__get_TMPDIR;
    # \Q..\E here is due the bug:
    # Nested quantifiers in regex; marked by <-- HERE in m/^/home/repocop/mlt++ <-- HERE -0.2.2-alt2.9666//
    $relativeold=~s!^\Q$tmpdir\E/!!;
    $relativenew=~s!^\Q$tmpdir\E/!!;
    if (substr($outfile,0,1) ne '/') {
	my $curdir=`pwd`;
	chomp $curdir;
	$outfile=$curdir.'/'.$outfile;
    }
    $class->__chtmpdir_and_call("diff -u $relativeold $relativenew $direction $outfile");
}

sub __chtmpdir_and_call {
    my $class=shift;
    my $cmd=shift;
    my $curdir=`pwd`;
    chomp $curdir;
    chdir $class->__get_TMPDIR || warn "can't chdir to $class->__get_TMPDIR";
    system ($cmd); # && warn "__chtmpdir_and_call:in $class->__get_TMPDIR: command failed: $cmd";
    chdir $curdir;
}

my $__tmpdir;

sub __get_TMPDIR {
    my $class=shift;
    return $__tmpdir;
}

sub __mk_TMPDIR {
    my $self=shift;
    if (!defined $__tmpdir || -d $__tmpdir) {
	$__tmpdir = tempdir( "repocop-fix.XXXXXXXX", TMPDIR => 1, CLEANUP=>1);
	die "can't create temporary directory!" unless $__tmpdir;
    } else {
	mkpath([$__tmpdir]);
    }
}

sub __del_TMPDIR {
    if (defined $__tmpdir and -d $__tmpdir) {
	system 'chmod', '-Rf', 'u+rwX', $__tmpdir;
	rmtree ([$__tmpdir]);
	undef $__tmpdir;
    }
}

#--------------------------
sub load_file {
    my ($name)=@_;
    open (my $fh, '<', $name) || die "can't open $name: $!";
    local $/; # enable localized slurp mode
    my $content = <$fh>;
    close ($fh) || die "can't close $name: $!";
    return $content;
}

sub write_file {
    my ($name, $str, $mode)=@_;
    $mode ||= '>';
    open (my $fd, $mode, $name) or die "can't open $name: $!";
    print $fd $str;
    close ($fd) || die "can't close $name: $!";
}

1;
