package RPM::Source::Transformation::Factory::CommandLine;

use strict;
use warnings;
use Carp;

use RPM::Source::Tools::Uupdate;
# DEPENDS on @datapath and @patchdir resolution.
use RPM::Source::Transformation::Resource::Datapath;
use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our $VERSION=0.003;

our (@opt_copy_to_sources, @opt_spec_apply_patch,@opt_add_source);
my (@opt_macrodef,@opt_macroundef,$opt_uupdate,$opt_rename,$opt_repack);

our $verbose=1;

sub longopt_usage {
    print "  Editing options:
	--macrodef '<name> <value>' set the value of an rpm macro.
	--macroundef <name> undefine rpm macro.
	  It does not change spec but influence spec processing.
	--add-source <file> (can be specified multiple times)
	--copy_to_sources <file> (can be specified multiple times)
	--copy-to-sources <file> - alias for --copy_to_sources
	--rename 'new name of source rpm package'
	--spec_apply_patch <patch file> -- apply <patch file> to spec
	--repack   -- repack src.rpm archives to plain tar (saves space with native xz)
	--uupdate  -- perform a version update using a source watch file
";
}

our @LONGOPT=(
    'macrodef=s'	=> \@opt_macrodef,
    'macroundef=s'	=> \@opt_macroundef,
    'add-source=s'	=> \@opt_add_source,
    'copy_to_sources=s'	=> \@opt_copy_to_sources,
    'copy-to-sources=s'	=> \@opt_copy_to_sources,
    'spec_apply_patch=s' => \@opt_spec_apply_patch,
    'rename=s'	=> \$opt_rename,
    'repack'	=> \$opt_repack,
    'uupdate'	=> \$opt_uupdate,
);

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-macrodef => \@opt_macrodef,
	-macroundef => \@opt_macroundef,
	-rename => $opt_rename,
	-repack => $opt_repack,
	-uupdate => $opt_uupdate,
    });
    $extopt->{-opt_add_source}=\@opt_add_source if @opt_add_source;
    $extopt->{-opt_copy_to_sources}=\@opt_copy_to_sources if @opt_copy_to_sources;
    $extopt->{-opt_spec_apply_patch}=\@opt_spec_apply_patch if @opt_spec_apply_patch;
}

sub transformations {
    return 
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_AT_BEGIN,
	    -name => 'CommandLine define macro',
	    -opt_handler_sub=>\&opt_handler_macro_AT_BEGIN),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_PRECONVERT,
	    -name => 'CommandLine uupdate',
	    -opt_handler_sub=>\&opt_handler_uupdate_PRECONVERT),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_PRECONVERT,
	    -name => 'CommandLine tarball repack',
	    -opt_handler_sub=>\&opt_handler_repack_PRECONVERT),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_POSTCONVERT,
	    -name => 'CommandLine rename',
	    -opt_handler_sub=>\&opt_handler_rename_POSTCONVERT),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_COMMANDLINE,
	    -name => 'CommandLine main body',
	    -opt_handler_sub=>\&opt_handler);
}

sub opt_handler_macro_AT_BEGIN {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    my $macrodef=$opt->{-macrodef};
    my $macroundef=$opt->{-macroundef};
    if ($macrodef) {
	for my $val (@$macrodef) {
	    $val =~ /^([^\s=]+)[\s=]+(.*)$/;
	    $macros->set_raw($1,$2);
	}
    }
    if ($macroundef) {
	for my $val (@$macroundef) {
	    $macros->del_raw($val);
	}
    }
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

sub opt_handler_rename_POSTCONVERT {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    my $_rename;
    $_rename=$opt->{-rename} if $opt->{-rename} and $opt->{-rename} ne 'none';
    $_rename=$spec->macros->macro_subst($_rename) if $_rename;
    $spec->rename_main_package($_rename) if $_rename;
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

sub opt_handler_uupdate_PRECONVERT {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    if ($opt->{-uupdate}) {
	my $death_flag=1;
	my $new_version=&RPM::Source::Tools::Uupdate::uupdate($spec, $death_flag);
	if (defined $new_version) {
	    $opt->{-version} = $new_version;
	} else {
	    return TRANSFORMATION_FAILURE;
	}
    }
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

sub opt_handler_repack_PRECONVERT {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    if ($opt->{-repack}) {
	&RPM::Source::Tools::Uupdate::repack($spec);
    }
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}


sub opt_handler {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    my $macros=$spec->macros();
    my $patchdirref=$resourceobj->patchdirref();
    &__::DTransform::__add_source($spec,$opt->{-opt_add_source},$patchdirref);
    &__::DTransform::__copy_to_sources($spec,$opt->{-opt_copy_to_sources},$patchdirref);
    &__::DTransform::__spec_apply_patch($spec,$opt->{-opt_spec_apply_patch},$patchdirref);
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

# private Transform modules
package __::DTransform;

use strict;
use warnings;

sub __find_file_in_dirs {
    my ($file, $dirsref, $message_title)=@_;
    die "Oops: internal: file or dirsref is not defined!" if not defined $file or not $dirsref or ref $dirsref ne 'ARRAY';
    my @found=grep {-e $_} $file, map {"$_/$file"} @$dirsref;
    die "$message_title: $file not found\n" if (!@found);
    return $found[0];
}

sub __add_source {
    my ($spec, $fileref, $dirsref)=@_;
    return unless $fileref;
    die "Oops: internal: add_source is not an array!" if ref $fileref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$fileref) {
	$spec->add_source(&__find_file_in_dirs($file,$dirsref,'add_source'));
    }
}

sub __copy_to_sources {
    my ($spec, $fileref, $dirsref)=@_;
    return unless $fileref;
    die "Oops: internal: copy_to_sources is not an array!" if ref $fileref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$fileref) {
	$spec->copy_to_sources(&__find_file_in_dirs($file,$dirsref,'copy_to_sources'));
    }
}

sub __spec_apply_patch {
    my ($spec, $fileref, $dirsref)=@_;
    return unless $fileref;
    die "Oops: internal: spec_apply_patch is not an array!" if ref $fileref ne 'ARRAY';
    foreach my $file (map {$spec->macros()->macro_subst($_)} @$fileref) {
	$spec->spec_apply_patch(
	    SILENT => $verbose ? 0:1,
	    PATCHFILE=>&__find_file_in_dirs($file,$dirsref,'spec_apply_patch'),
	    );
    }
}


1;
