package RPM::Source::Transformation::Factory::Hook;

use strict;
use warnings;
use Carp;
use Cwd;

# DEPENDS on @datapath and @patchdir resolution.
use RPM::Source::Transformation::Resource::Datapath;
use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our ($opt_no_default_hook);
our (@hooks);
our $verbose=1;

our @COMMON_SPECHOOKS;
our @COMMON_PREHOOKS;

sub longopt_usage {
    print "
  Hook load options:
	--hook /path/to/<hook code> - load a subroutine hook.
	  Multiple --hook options are allowed.
	--no-default-hook do not look for the default hook '\%{name}.pl'
";
}

our @LONGOPT=(
    'no-default-hook|no-defaulthook'	=> sub {$opt_no_default_hook=1},
    'default-hook'	=> sub {$opt_no_default_hook=0},
    'hook=s'	=> \@hooks,
);

sub process_options {
    my ($self)=@_;
    if (@hooks) {
	&SpecHook::Loader::load_hooks(@hooks);
	push @COMMON_PREHOOKS, @SpecHook::Loader::PREHOOKS;
	push @COMMON_SPECHOOKS, @SpecHook::Loader::SPECHOOKS;
	@hooks=();
    }	
}

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-no_default_hook => $opt_no_default_hook,
    })
}

sub transformations {
    my (undef,$opt,$resourceobj)=@_;
    if ($opt->{-no_default_hook}) {
	print STDERR "DEBUG: HookFactory: -no_default_hook enabled, hook search skipped\n" if $verbose>1;
    } else {
	@hooks=$resourceobj->find_named_filesystem_resources('hooks','.pl');
    }

    ##### OOPS!!! move them (@PREHOOKS,@SPECHOOKS) in opts ... 
    # they are state of this factory instance or OPTS instance...
    my (@PREHOOKS,@SPECHOOKS);
    if (@hooks) {
	if ($RPM::Source::Transform::verbose) {
	    my $cwd=cwd();
	    my @messages=@hooks;
	    map {s!^${cwd}/?!!} @messages;
	    map {print STDERR "loading default hook $_\n"} @messages;
	}	
	&SpecHook::Loader::load_hooks(@hooks);
	@hooks=();
	@PREHOOKS=@SpecHook::Loader::PREHOOKS;
	@SPECHOOKS=@SpecHook::Loader::SPECHOOKS;
    }
    return 
	RPM::Source::Transformation::HookWrapper->new(
	-name => 'prehooks',
	-priority => PRIORITY_PREHOOKS,
	-hookref => [@COMMON_PREHOOKS, @PREHOOKS], 
	),
	RPM::Source::Transformation::HookWrapper->new(
	-name => 'user hooks',
	-priority => PRIORITY_HOOKS,
	-hookref => [@COMMON_SPECHOOKS, @SPECHOOKS], 
	),
}

package RPM::Source::Transformation::HookWrapper;

use strict;
use warnings;
use Carp;

sub new {
    my $class = shift;
    my $self = {
	-priority => 0,
	@_
    };
    die "new: -hookref is required" unless $self->{-hookref};
    bless $self, $class;
    return $self;
}

sub apply {
    my ($self,$spec,$parent,$opt,$resourceobj)=@_;
    die "$spec is undefined" unless $spec;
    &_apply_hooks($spec,$parent,$self->{-hookref});
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

sub _apply_hooks {
    my ($spec,$parent, $hookref)=@_;
    if (@$hookref) {
	$spec->{WARN_NOTAPPLIED}++;
	foreach my $code (@$hookref) {&$code($spec,$parent)}
	$spec->{WARN_NOTAPPLIED}--;
    }
}


1;

package SpecHook::Loader;

BEGIN {
    use vars qw/@SPECHOOKS @PREHOOKS/;
}

sub load_hooks {
    my (@hooks)=@_;
    @PREHOOKS=();
    @SPECHOOKS=();
    my @SAVE_INC=%INC;
    foreach my $hook (@hooks) {
	eval {
	    print STDERR "loading hook $hook\n" if $RPM::Source::Transform::verbose>1;
	    require $hook;
	};
	if ($@ && $hook!~/\.pl$/) {
	    eval {
		print STDERR "loading hook $hook.pl\n" if $RPM::Source::Transform::verbose>1;
		require $hook.'.pl';
	    };
	}
	die "can't load hook $hook: $@\n" if $@;
	# TODO: no sense for multiple hooks
	die "script $hook does not have a valid hook!\n" unless @PREHOOKS or @SPECHOOKS;
    }
    # othervise shared hooks will fail to load next time.
    %INC=@SAVE_INC;
}

1;
