package RPM::Source::Transformation::Factory::RaiseRelease;

use strict;
use warnings;
use Cwd;
use Carp;

use RPM::Source::Transformation::Factory;
our @ISA=qw/RPM::Source::Transformation::Factory/;

push @RPM::Source::TransformContainer::FACTORY_CLASSES, __PACKAGE__;

our $VERSION=0.04;
our $verbose=1;

# opt'ified, seems to be not used
our ($opt_default_packager_policy,
     $opt_version,$opt_release,$opt_epoch,$opt_serial,
);

#### used in jppimport
our $packager;

# transform mode: parent is igonored at release increment,
our $force_transform_mode; # changelog is from origin, not from parent
our $nextrel='nmu';

# NOTE: $defaultchangelog is appenable default changelog.
# hooks can append entries to the default changelog
# whereas $changelog_entry will override any of them.
# NOTE: 2 possible entry points for default changelog:
# 1) new(CHANGELOG=>, 2) set_default_changelog/add_default_changelog
# TODO: rename CHANGELOG => -changelog in R::S::E? used in repocop
our ($opt_changelog_packager,$opt_changelog_entry,$opt_changelog_date);
# end opt'ified
our $defaultchangelog='- ';
our $version_comparator=\&__::Transform::__rpm_version_compare;

# optified
my $opt_refresh_timestamp;

# should not be optified; overrides supplied values.
my $usersetdefaultchangelog;

our %NEXTREL_METHOD=(
    'incr' => \&__::Transform::nextrel_incr,
    'incrfirst' => \&__::Transform::nextrel_incrfirst,
    'incrsecond' => \&__::Transform::nextrel_incrsecond,
    'qa' => \&__::Transform::nextrel_qa,
    'nmu' => \&__::Transform::nextrel_nmuappend,
    'nmuappend' => \&__::Transform::nextrel_nmuappend,
    'nmuadd' => \&__::Transform::nextrel_nmuadd,
    'none' => \&__::Transform::nextrel_none,
    );

our %INITREL_METHOD=(
    );

our %REL_METHOD_DESCR=(
    'incr' => 'increment release digits.',
    'incrfirst' => 'increment first component of release number, keeping the tail.',
    'incrsecond' => 'increment the last component of release number.',
    'qa' => 'add/increment .qaN',
#    'nmu' => '',
    'nmuappend' => 'append .1',
    'nmuadd' => 'append .1 for a single-component release or increment the second component.',
    'none' => 'do nothing.',
#    '' => '',
    );

sub _print_nextrel_methods {
    print "	Available next release methods: 
";
    my %main_name=map {$NEXTREL_METHOD{$_}=>$_} keys (%REL_METHOD_DESCR);
    my %synonym=map {$_=>[]} keys (%REL_METHOD_DESCR);
    foreach my $sname (sort keys(%NEXTREL_METHOD)) {
	my $description=$REL_METHOD_DESCR{$sname};
	my $mname=$main_name{$NEXTREL_METHOD{$sname}};
	push@{$synonym{$mname}}, $sname if $mname and not defined $description;
    }
    foreach my $sname (sort keys(%NEXTREL_METHOD)) {
	my $description=$REL_METHOD_DESCR{$sname};
	if (defined $description) {
	    print "\t\t$sname\t- $description\n";
	    my @synonyms = @{$synonym{$sname}};
	    print "\t\t\t  Synonyms: ".join(',',@synonyms)."\n" if @synonyms;
	}
    }
}

sub register_nextrel_method {
    my $in = {@_};
    my $name=$in->{-name};
    warn "register_nextrel_method: $name already registered" if $NEXTREL_METHOD{$name};
    my $nextrel=$in->{-nextrel_method};
    my $initrel=$in->{-initrel_method};
    $NEXTREL_METHOD{$name}=$nextrel if $nextrel;
    $INITREL_METHOD{$name}=$initrel if $initrel;
    $REL_METHOD_DESCR{$name}=$in->{-description};
    die "Oops: method registration for $name failed: no method" if not $nextrel and not $initrel;
}

sub longopt_usage {
    print "
  Release options:
	--tag-packager-replace 'new value' set Packager: tag to a new value.
	  special values are: 
	    auto - use rpm's \%{packager} value
	    none - clear tag
	    parent - use parent's value of Packager:
	--tag-packager-default 'new value' - if the Packager: tag is not defined
	  set it to the new value. This option allows comma separated list
	  of multiple sources, for example, 'parent,A. R. <ar\@b.org>'
	--changelog '- changelog entry'
	--serial 'new value'
	--epoch 'new value'
	--version 'version' - set new package version
	--release 'release' - if we need to adjust the package release
	--nextrel|--next-release-policy <policy>
	  known next release policies: 
	".join(', ',sort(keys %NEXTREL_METHOD))."
	  default policy is $nextrel.
	--refresh-timestamp update timestamp in existing changelog
";
}

our @LONGOPT=(
    'ch|changelog=s'   => \$usersetdefaultchangelog,
    # changelog-override options
    'changelog-entry=s'	=> \$opt_changelog_entry,
    'changelog-packager=s'   => \$opt_changelog_packager,
    'changelog-date=s'	=> \$opt_changelog_date,
    'epoch=s'   => \$opt_epoch,
    'nextrel|next-release-policy=s' => \$nextrel,
    'refresh-timestamp' => \$opt_refresh_timestamp,
    'release=s'	=> \$opt_release,
    'serial=s'	=> \$opt_serial,
    'version=s'	=> \$opt_version,
    'tag-packager-default=s' => \$opt_default_packager_policy,
    'tag-packager-replace=s' => \$packager,
    'packager=s' => \$packager,
);

sub register_messages {
    my ($self,$extopt)=@_;
    ### TODO: use some fixed keys hash here
    RPM::Source::Transformation::Component::add_opt($extopt, {
	-packager => $packager,
	-default_packager_policy => $opt_default_packager_policy,
	-default_changelog => $defaultchangelog,
	-version => $opt_version,
	-release => $opt_release,
	-serial => $opt_serial,
	-epoch => $opt_epoch,
	-nextrel => $nextrel,
	-refresh_timestamp => $opt_refresh_timestamp,
	-changelog_entry => $opt_changelog_entry,
	-changelog_packager => $opt_changelog_packager,
	-changelog_date => $opt_changelog_date,
	});
}

sub transformations {
    return 
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_AT_BEGIN,
	    # used in CheckIfChangedPlayer
	    -name => 'init default changelog',
	    -opt_handler_sub=>\&opt_handler_init_default_changelog_AT_BEGIN),
	RPM::Source::Transformation::OptHandlerWrapper->new(
	    -priority => PRIORITY_AT_END,
	    # used in CheckIfChangedPlayer
	    -name => 'raise release',
	    -opt_handler_sub=>\&opt_handler);
}

sub opt_handler_init_default_changelog_AT_BEGIN {
    my (undef,$spec,$parent,$opt)=@_;
    my $def_changelog=$opt->{-default_changelog};
    if ($usersetdefaultchangelog) {
	$spec->set_default_changelog($usersetdefaultchangelog);
    } elsif ($def_changelog) {
	$spec->set_default_changelog($def_changelog);
    }
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

# there is a reference to this subroutine in repocop; TODO: refactor it
sub opt_handler {
    my (undef,$spec,$parent,$opt)=@_;
    my $version=$opt->{-version};
    my ($nextrel,$force_transform_mode)=($opt->{-nextrel},$opt->{-force_transform_mode});

    &__::Transform::__set_packager	($spec,$parent,$opt->{-default_packager_policy},$opt->{-packager});
    &__::Transform::__refresh_timestamp($spec) if $opt->{-refresh_timestamp};

    # used in version comparator
    my ($oldname,$newname);
    $oldname=$parent->main_section->get_tag('Name') if $parent;
    $newname=$spec->main_section->get_tag('Name');

    my $new_serial=&__::Transform::__set_serial	($spec,$parent,$version,$oldname,$newname,$opt->{-serial},$opt->{-epoch});
    unless ($nextrel eq 'none' and not defined $version) {
	my @old_evr=$spec->_get_evr();
	my @new_evr=@old_evr; $new_evr[0]=$new_serial;
	$parent=&__::Transform::__set_version($spec,$parent,$version,\@new_evr) if defined $version;
	&__::Transform::__increment_release($spec,$parent,$oldname,$newname,$nextrel,$opt->{-release},$force_transform_mode,\@new_evr);
	&__::Transform::__set_changelog	($spec,$parent,$opt->{-changelog_entry},$opt->{-changelog_packager},$opt->{-changelog_date},$force_transform_mode);
        # TODO: update %release, %version, %serial macros to final evaled values.
        # TODO: a solution for 'fixes ahead' - %define release, %define version

	&__::Transform::__rename_sources_and_patches($spec,$parent,\@old_evr,\@new_evr);
    }
    return RPM::Source::Transformation::Factory::TRANSFORMATION_OK;
}

# private Transform modules
package __::Transform;

use strict;
use warnings;
use File::Basename;
use RPM qw/vercmp evrcmp/;
use File::Temp qw/tempfile tempdir/;

# TODO:
sub __rpm_version_compare {
    my ($oldname,$newname,$oldversion,$newversion)=@_;
    return RPM::evrcmp($oldversion,$newversion);
}


sub __refresh_timestamp {
    my ($spec) = @_;
    my $changelogsec=$spec->get_section('changelog');
    die "Refresh timestamp: no changelog found" unless $changelogsec;
    my $date=`LC_TIME=C date '+%a %b %d %Y'`;
    chomp $date;
    warn "changelog update failed" unless $changelogsec->get_bodyref
	->[1]=~s/^\* \w\w\w \w\w\w \d\d 20\d\d /* $date /;
}

sub __set_packager {
    my ($spec,$parent,$default_packager_policy,$packager) = @_;
    my $mainsec=$spec->main_section();
    my $packager_tag=$mainsec->get_tag('Packager');
    if ($packager) {
	if ($packager eq 'auto') {
	    $packager=$spec->macros()->macro_subst('%{packager}');
	} elsif ($packager eq 'none') {
	    $mainsec->clear_tag('Packager');
	} elsif ($packager eq 'parent') {
	    die "--packager=parent but there is no parent\n" unless $parent;
	    my $candidate=$parent->main_section()->get_tag('Packager');
	    if ($candidate) {
		$mainsec->set_tag('Packager',$candidate);
	    } else {
		$mainsec->clear_tag('Packager');
	    }
	} else {
	    $mainsec->set_tag('Packager',$packager);
	}
    } elsif (! $packager_tag and $default_packager_policy) {
	foreach my $policy (split(/,/,$default_packager_policy)) {
	    if ($policy eq 'parent') {
		next unless $parent;
		my $candidate=$parent->main_section()->get_tag('Packager');
		if ($candidate) {
		    $mainsec->set_tag('Packager',$candidate);
		    last;
		}
	    } elsif ($policy eq 'auto') {
		my $candidate=$spec->macros()->macro_subst('%{packager}');
		if ($candidate) {
		    $mainsec->set_tag('Packager',$candidate);
		    last;
		}
	    } elsif ($policy =~ m'\@') {
		# default packager
		$mainsec->set_tag('Packager',$policy);
		last;
	    }
	}
    }
}

sub __set_changelog {
    my ($spec,$parent,$changelog_entry,$changelog_packager,$changelog_date,$force_transform_mode)=@_;
    my $changelogsec=$spec->get_section('changelog');
    die "Oops! changelog section not found" unless $changelogsec;
    if ($parent && !$force_transform_mode) {
	$changelogsec->set_body($parent->get_section('changelog')->get_bodyref());
    }
    # old hack to re-split lines against broken add_changelog entries
    $changelogsec->set_body($changelogsec->get_bodyref());

    my @args;
    push @args, 'ENTRY' => $changelog_entry if $changelog_entry;
    push @args, 'PACKAGER' => $changelog_packager if $changelog_packager;
    push @args, 'DATE' => $changelog_date if $changelog_date;
    $spec->add_changelog(@args);
}

sub __serial_val {
    my ($epoch,$serial) =@_;
    if (defined $epoch) {
	if (defined $serial) {
	    return ($epoch > $serial ) ? $epoch:$serial;
	} else {
	    return $epoch;
	}
    } else {
	if (defined $serial) {
	    return $serial;
	} else {
	    return $epoch;
	}
    }
    return;
}

sub __set_serial {
    my ($spec,$parent,$version,$oldname,$newname,$serial,$epoch)=@_;
    my $specmain=$spec->main_section();
    if ($parent and not defined $epoch and not defined $serial) {
	my $parentmain=$parent->main_section();
	my $parentepoch = $parentmain->get_tag('Epoch');
	my $parentserial = $parentmain->get_tag('Serial');
	my $oldepoch = $specmain->get_tag('Epoch');
	my $oldserial = $specmain->get_tag('Serial');
	my $parentval= &__serial_val($parentepoch,$parentserial);
	my $oldval= &__serial_val($oldepoch,$oldserial);
	if (! defined $oldval || (defined $parentval and $parentval > $oldval)) {
	    if (defined $parentepoch) {
		$epoch=$parentepoch;
		$oldepoch=$parentepoch;
	    }
	    if (defined $parentserial) {
		$serial=$parentserial;
		$oldserial=$parentserial;
	    }
	    $oldval=$parentval;
	    print STDERR "note: picked up serial $parentserial from parent\n" if defined($parentserial) and $verbose;
	    print STDERR "note: picked up epoch $parentepoch from parent\n" if defined($parentepoch) and $verbose;
	}
	my $specver = $spec->main_section()->get_tag('Version');
	$specver=$version if defined($version);
	my $parentver = $parent->main_section()->get_tag('Version');
	my $vercmp = &$RPM::Source::Transformation::Factory::RaiseRelease::version_comparator($oldname, $newname, $parentver,$specver);
	if ($vercmp > 0 and (! defined $oldval || (defined $parentval and $parentval >= $oldval))) {
	    print STDERR "DEBUG: downgrade: oldname=$oldname, newname=$newname, parentver=$parentver,specver=$specver cmp=$vercmp\n" if $verbose>1;
	    if (defined($oldserial)) {
		$serial=++$oldserial;
		print STDERR "note: downgrade detected; increasing serial to $serial\n" if $verbose;
	    } elsif (defined($oldepoch)) {
		$epoch=++$oldepoch;
		print STDERR "note: downgrade detected; increasing epoch to $epoch\n" if $verbose;
	    } else {
		$serial=1;
		print STDERR "note: downgrade detected; increasing serial to $serial\n" if $verbose;
	    }
	}
    }
    my $retval;
    if (defined $epoch) {
	if ($epoch eq 'none') {
	    $specmain->exclude(qr'^Epoch:');
	    $spec->macros()->del_raw('epoch');
	} else {
	    $specmain->set_tag('Epoch',$epoch);
	    $spec->macros()->set_raw('epoch',$epoch);
	}
	$specmain->exclude(qr'^Serial:');
	$spec->macros()->del_raw('serial');
	$retval=$epoch;
    }
    if (defined $serial) {
	if ($serial eq 'none') {
	    $specmain->exclude(qr'^Serial:');
	    $spec->macros()->del_raw('serial');
	} else {
	    $specmain->set_tag('Serial',$serial);
	    $spec->macros()->set_raw('serial',$serial);
	}
	$specmain->exclude(qr'^Epoch:');
	$spec->macros()->del_raw('epoch');
	$retval=$serial;
    }

    # adjust epoch in strict dependencies
    my $has_epoch=defined $epoch;
    my $has_serial=defined $serial;
    if ($has_epoch or $has_serial) {
	foreach my $sec ($spec->get_sections) {
	    my $type=$sec->get_type;
	    next unless $type eq 'package';
	    if ($has_epoch) {
		$sec->map_body(
		sub {
		    if (/^Requires:/i) {
			s,\s+=\s+\%{version}-\%{release}, = %{?epoch:%epoch:}%{version}-%{release},g;
			s,\s+=\s+\%version-\%release, = %{?epoch:%epoch:}%version-%release,g;
		    }
		});
	    } elsif ($has_serial) {
		$sec->map_body(
		sub {
		    if (/^Requires:/i) {
			s,\s+=\s+\%{version}-\%{release}, = %{?serial:%serial:}%{version}-%{release},g;
			s,\s+=\s+\%version-\%release, = %{?serial:%serial:}%version-%release,g;
		    }
		});
	    }
	}
    }
    return $retval;
}

sub __set_version {
    my ($spec,$parent,$version,$new_spec_evr)=@_;
    return $parent unless defined $version;
    $parent=$spec->clone() unless defined $parent;
    my $mainsec=$spec->main_section();
    my $specver=$mainsec->get_tag('Version', RAW => 1);
    if ($specver eq '%version' or $specver eq '%{version}') {
	my $replaced=0;
	$mainsec->map_body(sub{$replaced||=s/^(\s*\%(?:define|global)\s+version\s+)\S+/$1$version/});
	unless ($replaced) {
	    print STDERR "WARNING:can't find the definition of \%version macro for ".$mainsec->get_tag('Name', RAW => 0)."\n";
	}
    }
    $mainsec->set_tag('Version', $version);
    $spec->macros->set_raw('version',$version);
    $new_spec_evr->[1]=$version;
    return $parent;
}

sub __increment_release() {
    my ($spec,$parent,$oldname,$newname,$nextrel_policy,$forced_value,$force_transform_mode,$new_spec_evr)=@_;
    if (!defined ($forced_value) && $parent && !$force_transform_mode) {
	my @parent_evr=$parent->_get_evr();
	my $vercmp = &$RPM::Source::Transformation::Factory::RaiseRelease::version_comparator($oldname, $newname, $parent_evr[1],$new_spec_evr->[1]);
	if ($vercmp < 0) {
	    # upgrade;
	    $forced_value='alt1';
	    my $initrel_method=$INITREL_METHOD{$nextrel_policy};
	    $forced_value=&$initrel_method($spec,$parent) if $initrel_method;
	    print STDERR "note: increment release: version upgrade: forced_value=$forced_value\n" if $verbose>1;
	} elsif ($vercmp>0) {
	    # downgrade;
	    # just increment release normally, it usually reasonable.
	    print STDERR "note: increment release: version downgrade.\n" if $verbose>1;
	}
    } 
    my $mainsec=$spec->main_section();
    my $macros=$spec->macros();
    my $bodyref=$mainsec->get_bodyref();
    for (my $i=0; $i<@$bodyref; $i++) {
	#if ($bodyref->[$i]=~/^(release:\s*)(\S+)(.+)/si) {
	# Release %macro arg contains spaces :(
	if ($bodyref->[$i]=~/^(release:\s*)(\S.+)$/si) {
	    my ($pre,$rel)=($1,$2);
	    $rel=~s/(\s*)$//;
	    my $post=$1;
	    $post="\n" unless $post;
	    # hack around;
	    # recurrent release like %define release alt1; Release %release.1.2.3
	    $rel=$macros->get_raw('release') if $rel =~ /\%{release}/ or $rel =~ /\%release/;
	    print STDERR "debug:before nextrel pre=[$pre] rel=[$rel]\n" if $verbose>2;
	    print STDERR "debug:forced_value=$forced_value\n" if $forced_value and $verbose>2;
	    $rel=$forced_value ? $forced_value : &__::Transform::__raw_increment_release($nextrel_policy, $rel,$spec,$parent);
	    $bodyref->[$i]=$pre.$rel.$post;
	    die "FATAL: no endline: [$pre][$rel][$post]" unless $bodyref->[$i]=~/\n$/;
	    $macros->set_raw('release',$macros->macro_subst($rel));
	    $new_spec_evr->[2]=$rel;
	    print STDERR "debug:after nextrel pre=[$pre] rel=[$rel]\n" if $verbose>2;
	}
    }
    $mainsec->set_body($bodyref);
}


sub __rename_sources_and_patches() {
    my ($spec,$parent,$old_evr,$new_evr)=@_;
    ## HACK to rename sources along with new release. 
    # DO not run in GEAR mode.
    return if $spec->{GEARDIR};

    my $old_vr=$old_evr->[1].'-'.$old_evr->[2];
    my $new_vr=$new_evr->[1].'-'.$new_evr->[2];

    print STDERR "__rename_sources_and_patches: old_vr=$old_vr new_vr=$new_vr\n" if $verbose>3;

    if ($spec->{SOURCERPM} and -d $spec->{SOURCEDIR} and $old_vr and $new_vr) {
	foreach my $file (glob "$spec->{SOURCEDIR}/*") {
	    next unless -f $file;
	    my $basename=basename($file);
	    my $copy=$basename;
	    $copy=~s/\Q$old_vr\E/$new_vr/;
	    if ($copy ne $basename) {
		system ('cp',$file,"$spec->{SOURCEDIR}/$copy")==0 or die "cp $file $spec->{SOURCEDIR}/$copy failed!";
	    }
	}

	my $prepsec=$spec->get_section('prep');
	#%setup -q -n %name-%version-%release
	if ($prepsec and $prepsec->match_body(qr'^\s*\%setup.+-n\s+.+-\%{?version}?-\%{?release}?')) {
	    my $repack=0;
	    foreach my $file (glob "$spec->{SOURCEDIR}/*") {
		next unless -f $file and $file=~/\.tar(\.(?:gz|bz2))?$/;
		my $arj='';
		if ($1) {
		    $arj='z' if ($1 eq '.gz');
		    $arj='j' if ($1 eq '.bz2');
		}
		my $tmpdir = tempdir( "tmp-unpack-source.XXXXXXXX", TMPDIR => 1, CLEANUP=>1);
		die "can't create temporary directory!" unless $tmpdir;
		#my $curdir=cwd;
		#chdir $tmpdir;
		system('tar','x'.$arj.'f',$file,'-C',$tmpdir)==0 or die 'tar x'.$arj."f $file -C $tmpdir failed!";
		my ($dir)=glob "$tmpdir/*";
		if ( -d $dir) {
		    my $basename=basename($dir);
		    my $copy=$basename;
		    $copy=~s/\Q$old_vr\E/$new_vr/;
		    if ($copy ne $basename) {
			system ('mv','-f',$dir,"$tmpdir/$copy")==0 or die "mv $dir $tmpdir/$copy failed!";
			system ('tar','c'.$arj.'f',$file,'-C',$tmpdir,$copy)==0 or die 'tar c'.$arj."f $file -C $tmpdir $copy failed!";
			$repack=1;
		    }
		}
	    }
	    unless ($repack) {
		print STDERR "WARNING: found \%setup -q -n ...-\%version-\%release\n";
	    }
	}

	# hack: fix version in patches
	my $mainsec=$spec->main_section();
	my $oldver=$old_evr->[1];
	my $oldrel=$old_evr->[2];
	my $newver=$new_evr->[1];
	my $newrel=$new_evr->[2];
	# vr's are already renamed; so no need to touch release

	print STDERR "__rename_sources_and_patches: oldver=$oldver oldrel=$oldrel newver=$newver newrel=$newrel\n" if $verbose>3;
	$mainsec->map_body(sub {
	    return unless /^((?:source|patch)[^:]*:\s+)(\S+)(\s*)$/i;
	    my $prefix=$1;
	    my $path=$2;
	    my $suffix=$3;
	    my $basename=basename($path);
	    my $url=substr($path,0,length($path)-length($basename));
	    # basename(macro_subst(basename()) helps for SourceN: %{someurl}file.tar
	    return if -e "$spec->{SOURCEDIR}/".basename($spec->macros()->macro_subst($basename));

	    my $new_filename=$basename;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    $new_filename=~s/\%\{?version\}?/$oldver/;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    $new_filename=~s/\%\{?version\}?/$newver/;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    $new_filename=~s/\Q$oldver\E/$newver/;
	    $new_filename=~s/\Q$oldrel\E/$newrel/;
	    return if &__check_if_exist_new_filename($spec,\$new_filename,$basename, $prefix.$url, $suffix);
	    warn "WARNING: setting new version: can't find $prefix$path. Pray for the best!";
			   });

    }
}

sub __check_if_exist_new_filename {
    my ($spec,$new_filenameref,$basename, $prefix,$suffix)=@_;
    my $expanded_name="$spec->{SOURCEDIR}/".basename($spec->macros()->macro_subst($$new_filenameref));
    if (-e $expanded_name) {
	$_=$prefix.$$new_filenameref.$suffix;
	return 1;
    }
    $$new_filenameref=~s/\.(?:gz|bz2|xz)$//;
    my $new_expanded_name=$expanded_name;
    $new_expanded_name=~s/\.(?:gz|bz2|xz)$//;
    if (-e $expanded_name.'.gz') {
	&run('gunzip',$expanded_name.'.gz');
    } elsif (-e $expanded_name.'.bz2') {
	&run('bunzip',$expanded_name.'.bz2');
    } elsif (-e $expanded_name.'.xz') {
	&run('unxz',$expanded_name.'.xz');
    }
    if (-e $new_expanded_name) {
	$_=$prefix.$$new_filenameref.$suffix;
	return 1;
    }

    $$new_filenameref=$basename;
    return 0;
}

sub __raw_increment_release {
    my ($nextrel_policy,$rel,$spec,$parent)=@_;
    $nextrel_policy||= 'incr';
    my $nextrel_method=$NEXTREL_METHOD{$nextrel_policy};
    unless ($nextrel_method) {
	&RPM::Source::Transformation::Factory::RaiseRelease::_print_nextrel_methods();
	die "Unknown next release policy: $nextrel_policy\n";
    }
    my $newrel = &{$nextrel_method}($rel,$spec,$parent);
    $newrel||= $rel.".1";
    return $newrel;
}


sub run {
    print STDERR "running: ", join(' ',@_),"\n" if $verbose > 4;
    system(@_)==0 or die "command failed: ".join(' ',@_);
}

#----------------------------------------------------------
#########################################
# private:  release update policies
#########################################
sub nextrel_incr {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~/^((?:\D+)?)(\d+)/) {
        my $relprefix=$1;
        my $digitrel=$2;
        $newrel = $relprefix.++$digitrel;
    }
    return $newrel;
}

sub nextrel_incrfirst {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~/^((?:\D+)?)(\d+)(.*)$/) {
        my $relprefix=$1;
        my $digitrel=$2; my $tail=$3;
        $newrel = $relprefix.++$digitrel.$tail;
    }
    return $newrel;
}

sub nextrel_incrsecond {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.(\d+)$//) {
	$newrel = $rel.".".($1+1);
    } else {
	$newrel = $rel.".1";
    }
    return $newrel;
}

sub nextrel_qa {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.qa(\d+)(?:\.(?:qa\d+|1))*$//) {
	$newrel = $rel.".qa".($1+1);
    } else {
	$newrel = $rel.".qa1";
    }
    return $newrel;
}

sub nextrel_nmuappend {
    my ($rel)=@_;
    return $rel.".1";
}

sub nextrel_nmuadd {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.(\d+)$//) {
	$newrel = $rel.".".($1+1);
    } else {
	$newrel = $rel.".1";
    }
    return $newrel;
}

sub nextrel_none {
    my ($rel)=@_;
    return $rel;
}


1;
